# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math
import sys
from types import SimpleNamespace

import pytest

from hypothesis import given, settings, strategies as st
from hypothesis.configuration import StorageDirectory
from hypothesis.internal.conjecture import providers
from hypothesis.internal.conjecture.choice import choice_equal
from hypothesis.internal.conjecture.providers import CONSTANTS_CACHE
from hypothesis.internal.constants_ast import Constants

from tests.common.debug import find_any
from tests.common.utils import Why, xfail_on_crosshair


# I tried using @given(st.integers()) here, but I think there is a bad interaction
# with CONSTANTS_CACHE when testing it inside of a hypothesis test.
@pytest.mark.parametrize("value", [2**20 - 50, 2**10 - 10, 129387123, -19827321, 0])
def test_can_draw_local_constants_integers(monkeypatch, value):
    # _get_local_constants normally invalidates this cache for us, but we're
    # monkeypatching it.
    CONSTANTS_CACHE.cache.clear()
    monkeypatch.setattr(
        providers, "_get_local_constants", lambda: Constants(integers={value})
    )
    find_any(st.integers(), lambda v: choice_equal(v, value))


@xfail_on_crosshair(Why.undiscovered)  # I think float_to_int is difficult for crosshair
@pytest.mark.parametrize("value", [1.2938, -1823.0239, 1e999, math.nan])
def test_can_draw_local_constants_floats(monkeypatch, value):
    CONSTANTS_CACHE.cache.clear()
    monkeypatch.setattr(
        providers, "_get_local_constants", lambda: Constants(floats={value})
    )
    find_any(st.floats(), lambda v: choice_equal(v, value))


@pytest.mark.parametrize("value", [b"abdefgh", b"a" * 50])
def test_can_draw_local_constants_bytes(monkeypatch, value):
    CONSTANTS_CACHE.cache.clear()
    monkeypatch.setattr(
        providers, "_get_local_constants", lambda: Constants(bytes={value})
    )
    find_any(st.binary(), lambda v: choice_equal(v, value))


@pytest.mark.parametrize("value", ["abdefgh", "a" * 50])
def test_can_draw_local_constants_string(monkeypatch, value):
    CONSTANTS_CACHE.cache.clear()
    monkeypatch.setattr(
        providers, "_get_local_constants", lambda: Constants(strings={value})
    )
    # we have a bunch of strings in GLOBAL_CONSTANTS, so it might take a while
    # to generate our local constant.
    find_any(
        st.text(),
        lambda v: choice_equal(v, value),
        settings=settings(max_examples=5_000),
    )


def test_actual_collection(monkeypatch, tmp_path):
    # covering test for doing some real work collecting constants. We'll fake
    # hypothesis as being the "local" module, just to get some real constant
    # collection going.

    # reset cache checks
    monkeypatch.setattr(providers, "_sys_modules_len", None)
    monkeypatch.setattr(providers, "_seen_modules", set())
    monkeypatch.setattr(providers, "is_local_module_file", lambda f: "hypothesis" in f)
    # Without monkeypatching this, this test only covers all the right lines the
    # first time it's executed, since it falls back to the cache the second time.
    # Reset to a guaranteed-empty storage directory to ensure consistent coverage.
    monkeypatch.setattr(
        "hypothesis.internal.constants_ast.storage_directory",
        lambda *names, **kwargs: StorageDirectory(
            tmp_path.joinpath(*names), home_directory=tmp_path
        ),
    )

    @given(st.integers())
    @settings(max_examples=100)
    def f(n):
        pass

    f()


def test_unhashable_sys_modules_entry(monkeypatch):
    # Regression test for https://github.com/HypothesisWorks/hypothesis/issues/4660
    # Some packages (e.g. cog) place unhashable objects like SimpleNamespace
    # in sys.modules. This should not crash _get_local_constants.
    monkeypatch.setattr(providers, "_sys_modules_len", None)
    monkeypatch.setattr(providers, "_seen_modules", set())
    monkeypatch.setitem(sys.modules, "_unhashable_test_mod", SimpleNamespace())

    providers._get_local_constants()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os
import re
import shlex
import subprocess
import sys

from hypothesistooling import ROOT


def print_command(command, args):
    args = list(args)
    ranges = []
    for i, v in enumerate(args):
        if os.path.exists(v):
            if not ranges or ranges[-1][-1] < i - 1:
                ranges.append([i, i])
            elif ranges[-1][-1] + 1 == i:
                ranges[-1][-1] += 1
    for i, j in ranges:
        if j > i:
            args[i] = "..."
            for k in range(i + 1, j + 1):
                args[k] = None
    args = [v for v in args if v is not None]
    print(command, *map(shlex.quote, args))


def run_script(script, *args, **kwargs):
    print_command(script, args)
    return subprocess.check_call([os.path.join(SCRIPTS, script), *args], **kwargs)


SCRIPTS = ROOT / "tooling" / "scripts"
COMMON = SCRIPTS / "common.sh"


def __calc_script_variables():
    exports = re.compile(r"^export ([A-Z_]+)(=|$)", flags=re.MULTILINE)

    common = COMMON.read_text(encoding="utf-8")

    for name, _ in exports.findall(common):
        globals()[name] = os.environ[name]


try:
    __calc_script_variables()
except Exception:
    pass


def tool_path(name):
    return os.path.join(os.path.dirname(sys.executable), name)


def pip_tool(name, *args, **kwargs):
    args = [str(arg) for arg in args]
    print_command(name, args)
    r = subprocess.call([tool_path(name), *args], **kwargs)

    if r != 0:
        sys.exit(r)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import json
import os
import re
import shutil
import subprocess
import sys

import requests
import tomli

import hypothesistooling as tools
from hypothesistooling import releasemanagement as rm

PACKAGE_NAME = "hypothesis-python"

HYPOTHESIS_PYTHON = tools.ROOT / PACKAGE_NAME
PYTHON_TAG_PREFIX = "hypothesis-python-"


BASE_DIR = HYPOTHESIS_PYTHON

PYTHON_SRC = HYPOTHESIS_PYTHON / "src"
PYTHON_TESTS = HYPOTHESIS_PYTHON / "tests"
DOMAINS_LIST = PYTHON_SRC / "hypothesis" / "vendor" / "tlds-alpha-by-domain.txt"

RELEASE_FILE = HYPOTHESIS_PYTHON / "RELEASE.rst"
RELEASE_SAMPLE_FILE = HYPOTHESIS_PYTHON / "RELEASE-sample.rst"

assert PYTHON_SRC.exists()


__version__ = None
__version_info__ = None

VERSION_FILE = os.path.join(PYTHON_SRC, "hypothesis/version.py")

with open(VERSION_FILE, encoding="utf-8") as fp:
    exec(fp.read())

assert __version__ is not None
assert __version_info__ is not None


def has_release():
    return RELEASE_FILE.exists()


def has_release_sample():
    return RELEASE_SAMPLE_FILE.exists()


def parse_release_file():
    return rm.parse_release_file(RELEASE_FILE)


def has_source_changes():
    return tools.has_changes([PYTHON_SRC])


def build_docs(*, builder="html", only=(), to=None):
    # See https://www.sphinx-doc.org/en/stable/man/sphinx-build.html
    tools.scripts.pip_tool(
        "sphinx-build",
        "--fail-on-warning",
        "--show-traceback",
        "--fresh-env",
        "--builder",
        builder,
        "docs",
        "docs/_build/" + (builder if to is None else to),
        *only,
        cwd=HYPOTHESIS_PYTHON,
    )


CHANGELOG_ANCHOR = re.compile(r"^\.\. _v\d+\.\d+\.\d+:$", flags=re.MULTILINE)
CHANGELOG_BORDER = re.compile(r"^-+$", flags=re.MULTILINE)
CHANGELOG_HEADER = re.compile(
    r"^\d+\.\d+\.\d+ - \d\d\d\d-\d\d-\d\d$", flags=re.MULTILINE
)


def update_changelog_and_version():
    global __version_info__
    global __version__

    contents = changelog()
    assert "\r" not in contents
    lines = contents.split("\n")
    for i, l in enumerate(lines):
        if CHANGELOG_ANCHOR.match(l):
            assert CHANGELOG_BORDER.match(lines[i + 2]), repr(lines[i + 2])
            assert CHANGELOG_HEADER.match(lines[i + 3]), repr(lines[i + 3])
            assert CHANGELOG_BORDER.match(lines[i + 4]), repr(lines[i + 4])
            assert lines[i + 3].startswith(
                __version__
            ), f"{__version__=}   {lines[i + 3]=}"
            beginning = "\n".join(lines[:i])
            rest = "\n".join(lines[i:])
            assert f"{beginning}\n{rest}" == contents
            break

    release_type, release_contents = parse_release_file()

    new_version_string, new_version_info = rm.bump_version_info(
        __version_info__, release_type
    )

    __version_info__ = new_version_info
    __version__ = new_version_string

    if release_type == "major":
        major, _, _ = __version_info__
        old = f"Hypothesis {major - 1}.x"
        beginning = beginning.replace(old, f"Hypothesis {major}.x")
        rest = "\n".join([old, len(old) * "=", "", rest])

    rm.replace_assignment(VERSION_FILE, "__version_info__", repr(new_version_info))

    heading_for_new_version = f"{new_version_string} - {rm.release_date_string()}"
    border_for_new_version = "-" * len(heading_for_new_version)

    new_changelog_parts = [
        beginning.strip(),
        "",
        f".. _v{new_version_string}:",
        "",
        border_for_new_version,
        heading_for_new_version,
        border_for_new_version,
        "",
        release_contents,
        "",
        rest,
    ]

    CHANGELOG_FILE.write_text("\n".join(new_changelog_parts), encoding="utf-8")

    # Replace the `since="RELEASEDAY"` argument to `note_deprecation`
    # with today's date, to record it for future reference.
    before = 'since="RELEASEDAY"'
    after = before.replace("RELEASEDAY", rm.release_date_string())
    for root, _, files in os.walk(PYTHON_SRC):
        for fname in (os.path.join(root, f) for f in files if f.endswith(".py")):
            with open(fname, encoding="utf-8") as f:
                contents = f.read()
            if before in contents:
                with open(fname, "w", encoding="utf-8") as f:
                    f.write(contents.replace(before, after))

    update_pyproject_toml()


def update_pyproject_toml():
    # manually write back these changes using regex instead of pulling in a
    # toml dependency for writing. tomli doesn't support writing, and
    # tomli-w doesn't support writing with comments.
    toml_p = HYPOTHESIS_PYTHON / "pyproject.toml"
    toml_data = tomli.loads(toml_p.read_text())
    extras = toml_data["project"]["optional-dependencies"]
    extras.pop("all")
    readme = (tools.ROOT / "README.md").read_text()
    content = toml_p.read_text()
    content = re.sub(
        r'\[project\.readme\].*content-type = "text/markdown"',
        f'[project.readme]\ntext = """{readme}"""\ncontent-type = "text/markdown"',
        content,
        flags=re.DOTALL,
    )

    all_extras = sorted(set(sum(extras.values(), [])))
    all_extras = json.dumps(all_extras).replace("\n", "\\n")
    content = re.sub(
        r"^all = \[.*\]$",
        f"all = {all_extras}",
        content,
        flags=re.MULTILINE,
    )
    toml_p.write_text(content)


CHANGELOG_FILE = HYPOTHESIS_PYTHON / "docs" / "changelog.rst"
DIST = HYPOTHESIS_PYTHON / "dist"


def changelog():
    return CHANGELOG_FILE.read_text(encoding="utf-8")


def build_distribution():
    if os.path.exists(DIST):
        shutil.rmtree(DIST)
    subprocess.check_output([sys.executable, "-m", "build", "--outdir", DIST])


def upload_distribution():
    tools.assert_can_release()

    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "twine",
            "upload",
            "--skip-existing",
            "--username=__token__",
            os.path.join(DIST, "*"),
        ]
    )

    # Construct plain-text + markdown version of this changelog entry,
    # with link to canonical source.
    build_docs(builder="text", only=["docs/changelog.rst"])
    textfile = os.path.join(
        HYPOTHESIS_PYTHON, "docs", "_build", "text", "changelog.txt"
    )
    with open(textfile, encoding="utf-8") as f:
        lines = f.readlines()
    entries = [i for i, l in enumerate(lines) if CHANGELOG_HEADER.match(l)]
    anchor = current_version().replace(".", "-")
    changelog_body = (
        "".join(lines[entries[0] + 2 : entries[1]]).strip()
        + "\n\n*[The canonical version of these notes (with links) is on readthedocs.]"
        f"(https://hypothesis.readthedocs.io/en/latest/changelog.html#v{anchor})*"
    )

    # Create a GitHub release, to trigger Zenodo DOI minting.  See
    # https://developer.github.com/v3/repos/releases/#create-a-release
    resp = requests.post(
        "https://api.github.com/repos/HypothesisWorks/hypothesis/releases",
        headers={
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {os.environ['GH_TOKEN']}",
            "X-GitHub-Api-Version": "2022-11-28",
        },
        json={
            "tag_name": tag_name(),
            "name": "Hypothesis for Python - version " + current_version(),
            "body": changelog_body,
        },
        timeout=120,  # seconds
    )

    # TODO: work out why this is 404'ing despite success (?!?) and fix it
    try:
        resp.raise_for_status()
    except Exception:
        import traceback

        traceback.print_exc()


def current_version():
    return __version__


def latest_version():
    versions = []

    for t in tools.tags():
        if t.startswith(PYTHON_TAG_PREFIX):
            t = t.removeprefix(PYTHON_TAG_PREFIX)
        else:
            continue
        assert t == t.strip()
        parts = t.split(".")
        assert len(parts) == 3
        v = tuple(map(int, parts))
        versions.append((v, t))

    _, latest = max(versions)

    return latest


def tag_name():
    return PYTHON_TAG_PREFIX + __version__


def get_autoupdate_message(domainlist_changed):
    if domainlist_changed:
        return (
            "This patch updates our vendored `list of top-level domains "
            "<https://www.iana.org/domains/root/db>`__,\nwhich is used by the "
            "provisional :func:`~hypothesis.provisional.domains` strategy.\n"
        )
    return (
        "This patch updates our autoformatting tools, "
        "improving our code style without any API changes."
    )



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math

import pytest

from hypothesis.extra.array_api import find_castable_builtin_for_dtype
from hypothesis.internal.floats import width_smallest_normals

from tests.array_api.common import dtype_name_params, flushes_to_zero
from tests.common.debug import (
    assert_all_examples,
    assert_no_examples,
    find_any,
    minimal,
)


@pytest.mark.parametrize("dtype_name", dtype_name_params)
def test_strategies_have_reusable_values(xp, xps, dtype_name):
    """Inferred strategies have reusable values."""
    strat = xps.from_dtype(dtype_name)
    assert strat.has_reusable_values


@pytest.mark.parametrize("dtype_name", dtype_name_params)
def test_produces_castable_instances_from_dtype(xp, xps, dtype_name):
    """Strategies inferred by dtype generate values of a builtin type castable
    to the dtype."""
    dtype = getattr(xp, dtype_name)
    builtin = find_castable_builtin_for_dtype(xp, xps.api_version, dtype)
    assert_all_examples(xps.from_dtype(dtype), lambda v: isinstance(v, builtin))


@pytest.mark.parametrize("dtype_name", dtype_name_params)
def test_produces_castable_instances_from_name(xp, xps, dtype_name):
    """Strategies inferred by dtype name generate values of a builtin type
    castable to the dtype."""
    dtype = getattr(xp, dtype_name)
    builtin = find_castable_builtin_for_dtype(xp, xps.api_version, dtype)
    assert_all_examples(xps.from_dtype(dtype_name), lambda v: isinstance(v, builtin))


@pytest.mark.parametrize("dtype_name", dtype_name_params)
def test_passing_inferred_strategies_in_arrays(xp, xps, dtype_name):
    """Inferred strategies usable in arrays strategy."""
    elements = xps.from_dtype(dtype_name)
    find_any(xps.arrays(dtype_name, 10, elements=elements))


@pytest.mark.parametrize(
    "dtype, kwargs, predicate",
    [
        # Floating point: bounds, exclusive bounds, and excluding nonfinites
        ("float32", {"min_value": 1, "max_value": 2}, lambda x: 1 <= x <= 2),
        (
            "float32",
            {"min_value": 1, "max_value": 2, "exclude_min": True, "exclude_max": True},
            lambda x: 1 < x < 2,
        ),
        ("float32", {"allow_nan": False}, lambda x: not math.isnan(x)),
        ("float32", {"allow_infinity": False}, lambda x: not math.isinf(x)),
        ("float32", {"allow_nan": False, "allow_infinity": False}, math.isfinite),
        # Integer bounds, limited to the representable range
        ("int8", {"min_value": -1, "max_value": 1}, lambda x: -1 <= x <= 1),
        ("uint8", {"min_value": 1, "max_value": 2}, lambda x: 1 <= x <= 2),
    ],
)
def test_from_dtype_with_kwargs(xp, xps, dtype, kwargs, predicate):
    """Strategies inferred with kwargs generate values in bounds."""
    strat = xps.from_dtype(dtype, **kwargs)
    assert_all_examples(strat, predicate)


def test_can_minimize_floats(xp, xps):
    """Inferred float strategy minimizes to a good example."""
    smallest = minimal(xps.from_dtype(xp.float32), lambda n: n >= 1.0)
    # TODO_IR should be resolved by float widths on the ir, see other TODO_IR comments
    assert smallest in {1, math.inf}


smallest_normal = width_smallest_normals[32]


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"min_value": -1},
        {"max_value": 1},
        {"min_value": -1, "max_value": 1},
    ],
)
def test_subnormal_generation(xp, xps, kwargs):
    """Generation of subnormals is dependent on FTZ behaviour of array module."""
    strat = xps.from_dtype(xp.float32, **kwargs).filter(lambda n: n != 0)
    if flushes_to_zero(xp, width=32):
        assert_no_examples(strat, lambda n: -smallest_normal < n < smallest_normal)
    else:
        find_any(strat, lambda n: -smallest_normal < n < smallest_normal)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import unicodedata

import pytest

from hypothesis import given, settings
from hypothesis.strategies import text


@pytest.mark.skipif(
    settings.get_current_profile_name() == "crosshair",
    reason="takes ~10 minutes; many-valued realization is slow",
)
@given(text(min_size=1, max_size=1))
@settings(max_examples=2000)
def test_does_not_generate_surrogates(t):
    assert unicodedata.category(t) != "Cs"



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import given, strategies as st


@st.composite
def strat(draw, x=0, /):
    return draw(st.integers(min_value=x))


@given(st.data(), st.integers())
def test_composite_with_posonly_args(data, min_value):
    v = data.draw(strat(min_value))
    assert min_value <= v


def test_preserves_signature():
    with pytest.raises(TypeError):
        strat(x=1)


def test_builds_real_pos_only():
    with pytest.raises(TypeError):
        st.builds()  # requires a target!



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

# -*- coding: utf-8 -*-
"""Statistical tests over the forms of the distributions in the standard set of
definitions.

These tests all take the form of a classic hypothesis test with the null
hypothesis being that the probability of some event occurring when
drawing data from the distribution produced by some specifier is >=
REQUIRED_P
"""

import collections
import math
import re

from hypothesis import HealthCheck, settings as Settings
from hypothesis.control import BuildContext
from hypothesis.errors import UnsatisfiedAssumption
from hypothesis.internal import reflection
from hypothesis.internal.conjecture.engine import ConjectureRunner
from hypothesis.strategies import (
    binary,
    booleans,
    floats,
    integers,
    just,
    lists,
    one_of,
    sampled_from,
    sets,
    text,
    tuples,
)

from tests.common.utils import no_shrink
from tests.conjecture.common import interesting_origin


class HypothesisFalsified(AssertionError):
    pass


def define_test(specifier, predicate, condition=None, p=0.5, suppress_health_check=()):
    runs = 100
    required_runs = int(runs * p)

    def run_test():
        if condition is None:

            def _condition(x):
                return True

            condition_string = ""
        else:
            _condition = condition
            condition_string = reflection.get_pretty_function_description(condition)
            condition_string = re.compile("^lambda[^:]*:\\s*").sub("", condition_string)

        def test_function(data):
            with BuildContext(data, wrapped_test=None):
                try:
                    value = data.draw(specifier)
                except UnsatisfiedAssumption:
                    data.mark_invalid()
                if not _condition(value):
                    data.mark_invalid()
                if predicate(value):
                    data.mark_interesting(interesting_origin())

        successes = 0
        actual_runs = 0
        for actual_runs in range(1, runs + 1):
            # We choose the max_examples a bit larger than default so that we
            # run at least 100 examples outside of the small example generation
            # part of the generation phase.
            runner = ConjectureRunner(
                test_function,
                settings=Settings(
                    max_examples=150,
                    phases=no_shrink,
                    suppress_health_check=suppress_health_check,
                ),
            )
            runner.run()
            if runner.interesting_examples:
                successes += 1
                if successes >= required_runs:
                    return

            # If we reach a point where it's impossible to hit our target even
            # if every remaining attempt were to succeed, give up early and
            # report failure.
            if (required_runs - successes) > (runs - actual_runs):
                break

        event = reflection.get_pretty_function_description(predicate)
        if condition is not None:
            event += "|"
            event += condition_string

        raise HypothesisFalsified(
            f"P({event}) ~ {successes} / {actual_runs} = "
            f"{successes / actual_runs:.2f} < {required_runs / runs:.2f}; "
            "rejected"
        )

    return run_test


test_can_produce_zero = define_test(integers(), lambda x: x == 0)
test_can_produce_large_magnitude_integers = define_test(
    integers(), lambda x: abs(x) > 1000
)
test_can_produce_large_positive_integers = define_test(integers(), lambda x: x > 1000)
test_can_produce_large_negative_integers = define_test(integers(), lambda x: x < -1000)


_factorials = {math.factorial(n) for n in range(9, 21)}


def is_factorial(n):
    return abs(n) in _factorials


test_can_produce_large_factorial = define_test(
    integers(), lambda n: n >= 50_000 and is_factorial(n), p=0.01
)
test_can_produce_above_large_factorial = define_test(
    integers(), lambda n: n >= 50_000 and is_factorial(n - 1), p=0.01
)
test_can_produce_below_large_factorial = define_test(
    integers(), lambda n: n >= 50_000 and is_factorial(n + 1), p=0.01
)
test_can_produce_large_factorial_negative = define_test(
    integers(), lambda n: n <= -50_000 and is_factorial(n), p=0.01
)
test_can_produce_above_large_factorial_negative = define_test(
    integers(), lambda n: n <= -50_000 and is_factorial(n - 1), p=0.01
)
test_can_produce_below_large_factorial_negative = define_test(
    integers(), lambda n: n <= -50_000 and is_factorial(n + 1), p=0.01
)


test_can_produce_unstripped_strings = define_test(text(), lambda x: x != x.strip())
test_can_produce_stripped_strings = define_test(text(), lambda x: x == x.strip())
# The pass probability here was previously 0.5, but some intermediate changes
# while working on the ir tweaked the distribution and made it flaky. We can
# reevaluate this once things have settled down, and likely bump the pass
# probability back up.
test_can_produce_multi_line_strings = define_test(text(), lambda x: "\n" in x, p=0.35)
test_can_produce_ascii_strings = define_test(
    text(), lambda x: all(ord(c) <= 127 for c in x)
)
test_can_produce_long_strings_with_no_ascii = define_test(
    text(min_size=5), lambda x: all(ord(c) > 127 for c in x), p=0.1
)
test_can_produce_short_strings_with_some_non_ascii = define_test(
    text(), lambda x: any(ord(c) > 127 for c in x), condition=lambda x: len(x) <= 3
)
test_can_produce_large_binary_strings = define_test(
    binary(), lambda x: len(x) > 10, p=0.3
)
test_can_produce_positive_infinity = define_test(floats(), lambda x: x == math.inf)
test_can_produce_negative_infinity = define_test(floats(), lambda x: x == -math.inf)
test_can_produce_nan = define_test(floats(), math.isnan)
test_can_produce_floats_near_left = define_test(floats(0, 1), lambda t: t < 0.2)
test_can_produce_floats_near_right = define_test(floats(0, 1), lambda t: t > 0.8)
test_can_produce_floats_in_middle = define_test(floats(0, 1), lambda t: 0.2 <= t <= 0.8)
test_can_produce_long_lists = define_test(
    lists(integers()), lambda x: len(x) >= 10, p=0.3
)
test_can_produce_short_lists = define_test(lists(integers()), lambda x: len(x) <= 10)
test_can_produce_the_same_int_twice = define_test(
    lists(integers()), lambda t: len(set(t)) < len(t)
)


def distorted_value(x):
    c = collections.Counter(x)
    return min(c.values()) * 3 <= max(c.values())


test_sampled_from_large_number_can_mix = define_test(
    lists(sampled_from(range(50)), min_size=50), lambda x: len(set(x)) >= 25
)
test_sampled_from_often_distorted = define_test(
    lists(sampled_from(range(5))), distorted_value, condition=lambda x: len(x) >= 3
)
test_non_empty_subset_of_two_is_usually_large = define_test(
    sets(sampled_from((1, 2))), lambda t: len(t) == 2
)
test_subset_of_ten_is_sometimes_empty = define_test(
    sets(integers(1, 10)), lambda t: len(t) == 0
)

test_mostly_sensible_floats = define_test(floats(), lambda t: t + 1 > t)
test_mostly_largish_floats = define_test(
    floats(), lambda t: t + 1 > 1, condition=lambda x: x > 0
)
test_ints_can_occasionally_be_really_large = define_test(
    integers(), lambda t: t >= 2**63
)

test_mixing_is_sometimes_distorted = define_test(
    lists(booleans() | tuples()),
    lambda x: distorted_value(map(type, x)),
    condition=lambda x: len(set(map(type, x))) == 2,
    suppress_health_check=[HealthCheck.filter_too_much],
)

test_mixes_2_reasonably_often = define_test(
    lists(booleans() | tuples()), lambda x: len(set(map(type, x))) > 1, condition=bool
)

test_partial_mixes_3_reasonably_often = define_test(
    lists(booleans() | tuples() | just("hi")),
    lambda x: 1 < len(set(map(type, x))) < 3,
    condition=bool,
)

test_mixes_not_too_often = define_test(
    lists(booleans() | tuples()), lambda x: len(set(map(type, x))) == 1, condition=bool
)

test_integers_are_usually_non_zero = define_test(integers(), lambda x: x != 0)
test_integers_are_sometimes_zero = define_test(integers(), lambda x: x == 0)
test_integers_are_often_small = define_test(integers(), lambda x: abs(x) <= 100)
test_integers_are_often_small_but_not_that_small = define_test(
    integers(), lambda x: 50 <= abs(x) <= 255
)


# This series of tests checks that the one_of() strategy flattens branches
# correctly.  We assert that the probability of any branch is >= 0.1,
# approximately (1/8 = 0.125), regardless of how heavily nested it is in the
# strategy.

# This first strategy chooses an integer between 0 and 7 (inclusive).
one_of_nested_strategy = one_of(
    just(0),
    one_of(
        just(1), just(2), one_of(just(3), just(4), one_of(just(5), just(6), just(7)))
    ),
)

for i in range(8):
    exec(f"""test_one_of_flattens_branches_{i} = define_test(
        one_of_nested_strategy, lambda x: x == {i}
    )""")


xor_nested_strategy = just(0) | (
    just(1) | just(2) | (just(3) | just(4) | (just(5) | just(6) | just(7)))
)

for i in range(8):
    exec(f"""test_xor_flattens_branches_{i} = define_test(
        xor_nested_strategy, lambda x: x == {i}
    )""")


# This strategy tests interactions with `map()`.  They generate integers
# from the set {1, 4, 6, 16, 20, 24, 28, 32}.
def double(x):
    return x * 2


one_of_nested_strategy_with_map = one_of(
    just(1),
    one_of(
        (just(2) | just(3)).map(double),
        one_of(
            (just(4) | just(5)).map(double),
            one_of((just(6) | just(7) | just(8)).map(double)),
        ).map(double),
    ),
)

for i in (1, 4, 6, 16, 20, 24, 28, 32):
    exec(f"""test_one_of_flattens_map_branches_{i} = define_test(
        one_of_nested_strategy_with_map, lambda x: x == {i}
    )""")


# This strategy tests interactions with `flatmap()`.  It generates lists
# of length 0-7 (inclusive) in which every element is `None`.
one_of_nested_strategy_with_flatmap = just(None).flatmap(
    lambda x: one_of(
        just([x] * 0),
        just([x] * 1),
        one_of(
            just([x] * 2),
            just([x] * 3),
            one_of(just([x] * 4), just([x] * 5), one_of(just([x] * 6), just([x] * 7))),
        ),
    )
)

for i in range(8):
    exec(f"""test_one_of_flattens_flatmap_branches_{i} = define_test(
        one_of_nested_strategy_with_flatmap, lambda x: len(x) == {i}
    )""")


xor_nested_strategy_with_flatmap = just(None).flatmap(
    lambda x: (
        just([x] * 0)
        | just([x] * 1)
        | (
            just([x] * 2)
            | just([x] * 3)
            | (just([x] * 4) | just([x] * 5) | (just([x] * 6) | just([x] * 7)))
        )
    )
)

for i in range(8):
    exec(f"""test_xor_flattens_flatmap_branches_{i} = define_test(
        xor_nested_strategy_with_flatmap, lambda x: len(x) == {i}
    )""")


# This strategy tests interactions with `filter()`.  It generates the even
# integers {0, 2, 4, 6} in equal measures.
one_of_nested_strategy_with_filter = one_of(
    just(0),
    just(1),
    one_of(just(2), just(3), one_of(just(4), just(5), one_of(just(6), just(7)))),
).filter(lambda x: x % 2 == 0)

for i in range(4):
    exec(f"""test_one_of_flattens_filter_branches_{i} = define_test(
        one_of_nested_strategy_with_filter, lambda x: x == 2 * {i}
    )""")


test_long_duplicates_strings = define_test(
    tuples(text(), text()), lambda s: len(s[0]) >= 5 and s[0] == s[1]
)
test_can_produce_nasty_strings = define_test(
    text(), lambda s: s in {"NaN", "Inf", "undefined"}, p=0.01
)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import assume, given, strategies as st
from hypothesis.internal.conjecture.junkdrawer import IntList

non_neg_lists = st.lists(st.integers(min_value=0, max_value=2**63 - 1))


@given(non_neg_lists)
def test_intlist_is_equal_to_itself(ls):
    assert IntList(ls) == IntList(ls)


@given(non_neg_lists, non_neg_lists)
def test_distinct_int_lists_are_not_equal(x, y):
    assume(x != y)
    assert IntList(x) != IntList(y)


def test_basic_equality():
    x = IntList([1, 2, 3])
    assert x == x
    t = x != x
    assert not t
    assert x != "foo"

    s = x == "foo"
    assert not s


def test_error_on_invalid_value():
    with pytest.raises(ValueError):
        IntList([-1])


def test_extend_by_too_large():
    x = IntList()
    ls = [1, 10**6]
    x.extend(ls)
    assert list(x) == ls



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from fractions import Fraction

import pytest

from hypothesis import (
    HealthCheck,
    Phase,
    assume,
    example,
    given,
    settings,
    strategies as st,
)
from hypothesis.errors import InvalidArgument
from hypothesis.internal.conjecture import utils as cu
from hypothesis.internal.conjecture.data import ConjectureData, Status, StopTest
from hypothesis.internal.coverage import IN_COVERAGE_TESTS

try:
    import numpy as np
except ImportError:
    np = None


def test_drawing_certain_coin_still_writes():
    data = ConjectureData.for_choices([True])
    assert data.draw_boolean(1)
    assert data.choices == (True,)


def test_drawing_impossible_coin_still_writes():
    data = ConjectureData.for_choices([False])
    assert not data.draw_boolean(0)
    assert data.choices == (False,)


def test_draws_extremely_small_p():
    data = ConjectureData.for_choices((True,))
    assert data.draw_boolean(0.5**65)


@example([Fraction(1, 3), Fraction(1, 3), Fraction(1, 3)])
@example([Fraction(1, 1), Fraction(1, 2)])
@example([Fraction(1, 2), Fraction(4, 10)])
@example([Fraction(1, 1), Fraction(3, 5), Fraction(1, 1)])
@example([Fraction(2, 257), Fraction(2, 5), Fraction(1, 11)])
@example([0, 2, 47])
@settings(
    deadline=None,
    suppress_health_check=list(HealthCheck),
    phases=[Phase.explicit] if IN_COVERAGE_TESTS else settings.default.phases,
)
@given(st.lists(st.fractions(min_value=0, max_value=1), min_size=1))
def test_sampler_distribution(weights):
    total = sum(weights)
    n = len(weights)

    assume(total > 0)

    probabilities = [w / total for w in weights]

    sampler = cu.Sampler(weights)

    calculated = [Fraction(0)] * n
    for base, alternate, p_alternate in sampler.table:
        calculated[base] += (1 - p_alternate) / n
        calculated[alternate] += p_alternate / n

    for expected, actual in zip(probabilities, calculated, strict=True):
        if isinstance(actual, Fraction):
            assert expected == actual
        else:
            assert abs(expected - actual) < 0.001


def test_sampler_does_not_draw_minimum_if_zero():
    sampler = cu.Sampler([0, 2, 47])
    assert sampler.sample(ConjectureData.for_choices([0, 0])) != 0


def test_sampler_shrinks():
    sampler = cu.Sampler([4.0, 8.0, 1.0, 1.0, 0.5])
    assert sampler.sample(ConjectureData.for_choices([0] * 3)) == 0


def test_can_force_sampler():
    sampler = cu.Sampler([0.5, 0.5])
    cd = ConjectureData.for_choices([0] * 100)
    assert sampler.sample(cd, forced=0) == 0
    assert sampler.sample(cd, forced=1) == 1


def test_combine_labels_is_distinct():
    x = 10
    y = 100
    assert cu.combine_labels(x, y) not in (x, y)


@given(st.integers())
def test_combine_labels_is_identity_for_single_argument(n):
    assert cu.combine_labels(n) == n


@pytest.mark.skipif(np is None, reason="requires Numpy")
def test_invalid_numpy_sample():
    with pytest.raises(InvalidArgument):
        cu.check_sample(np.array([[1, 1], [1, 1]]), "array")


@pytest.mark.skipif(np is None, reason="requires Numpy")
def test_valid_numpy_sample():
    cu.check_sample(np.array([1, 2, 3]), "array")


def test_invalid_set_sample():
    with pytest.raises(InvalidArgument):
        cu.check_sample({1, 2, 3}, "array")


def test_valid_list_sample():
    cu.check_sample([1, 2, 3], "array")


def test_choice():
    assert ConjectureData.for_choices([1]).choice([1, 2, 3]) == 2


def test_fixed_size_draw_many():
    many = cu.many(
        ConjectureData.for_choices([]), min_size=3, max_size=3, average_size=3
    )
    assert many.more()
    assert many.more()
    assert many.more()
    assert not many.more()


def test_astronomically_unlikely_draw_many():
    # Our internal helper doesn't underflow to zero or negative, but nor
    # will we ever generate an element for such a low average size.
    data = ConjectureData.for_choices((True,) * 1000)
    many = cu.many(data, min_size=0, max_size=10, average_size=1e-5)
    assert many.more()


def test_rejection_eventually_terminates_many():
    many = cu.many(
        ConjectureData.for_choices((True,) * 1000),
        min_size=0,
        max_size=1000,
        average_size=100,
    )
    count = 0

    while many.more():
        count += 1
        many.reject()

    assert count <= 100


def test_rejection_eventually_terminates_many_invalid_for_min_size():
    data = ConjectureData.for_choices((True,) * 1000)
    many = cu.many(data, min_size=1, max_size=1000, average_size=100)

    with pytest.raises(StopTest):
        while many.more():
            many.reject()

    assert data.status == Status.INVALID


def test_many_with_min_size():
    many = cu.many(
        ConjectureData.for_choices((False,) * 5),
        min_size=2,
        average_size=10,
        max_size=1000,
    )
    assert many.more()
    assert many.more()
    assert not many.more()


def test_many_with_max_size():
    many = cu.many(
        ConjectureData.for_choices((True,) * 5), min_size=0, average_size=1, max_size=2
    )
    assert many.more()
    assert many.more()
    assert not many.more()


def test_samples_from_a_range_directly():
    s = cu.check_sample(range(10**1000), "")
    assert isinstance(s, range)


def test_p_continue_to_average_saturates():
    assert cu._p_continue_to_avg(1.1, 100) == 100


def test_unhashable_calc_label():

    class Unhashable:
        def __call__(self):
            return None

        def __hash__(self):
            raise TypeError

    c1 = Unhashable()
    c2 = Unhashable()

    with pytest.raises(TypeError):
        assert cu.calc_label_from_hash(c1) == cu.calc_label_from_hash(c2)
    assert cu.calc_label_from_callable(c1) == cu.calc_label_from_callable(c2)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import enum
import functools
import itertools
import operator

import pytest

from hypothesis import given, settings, strategies as st
from hypothesis.errors import InvalidArgument
from hypothesis.internal.compat import bit_count
from hypothesis.strategies._internal.strategies import SampledFromStrategy

from tests.common.debug import find_any, minimal
from tests.common.utils import fails_with


@pytest.mark.parametrize("size", [100, 10**5, 10**6, 2**25])
@given(data=st.data())
def test_filter_large_lists(data, size):
    n_calls = 0

    def cond(x):
        nonlocal n_calls
        n_calls += 1
        return x % 2 != 0

    s = data.draw(st.sampled_from(range(size)).filter(cond))

    assert s % 2 != 0
    assert n_calls <= SampledFromStrategy._MAX_FILTER_CALLS


def rare_value_strategy(n, target):
    def forbid(s, forbidden):
        """Helper function to avoid Python variable scoping issues."""
        return s.filter(lambda x: x != forbidden)

    s = st.sampled_from(range(n))
    for i in range(n):
        if i != target:
            s = forbid(s, i)

    return s


@given(rare_value_strategy(n=128, target=80))
def test_chained_filters_find_rare_value(x):
    assert x == 80


@fails_with(InvalidArgument)
@given(st.sets(st.sampled_from(range(10)), min_size=11))
def test_unsat_sets_of_samples(x):
    raise AssertionError


@given(st.sets(st.sampled_from(range(50)), min_size=50))
def test_efficient_sets_of_samples(x):
    assert x == set(range(50))


class AnEnum(enum.Enum):
    a = enum.auto()
    b = enum.auto()


def test_enum_repr_uses_class_not_a_list():
    # The repr should have enough detail to find the class again
    # (which is very useful for the ghostwriter logic we're working on)
    lazy_repr = repr(st.sampled_from(AnEnum))
    assert lazy_repr == "sampled_from(tests.nocover.test_sampled_from.AnEnum)"


def test_repr_truncates_with_many_elements():
    s = st.sampled_from(list(range(10_000)))
    repr_limit = 512
    assert repr(s) == f"sampled_from([{', '.join(map(str, range(repr_limit)))}, ...])"


class AFlag(enum.Flag):
    a = enum.auto()
    b = enum.auto()
    c = enum.auto()


LargeFlag = enum.Flag("LargeFlag", {f"bit{i}": enum.auto() for i in range(64)})


class UnnamedFlag(enum.Flag):
    # Would fail under EnumCheck.NAMED_FLAGS
    a = 0
    b = 7


def test_flag_enum_repr_uses_class_not_a_list():
    lazy_repr = repr(st.sampled_from(AFlag))
    assert lazy_repr == "sampled_from(tests.nocover.test_sampled_from.AFlag)"


def test_exhaustive_flags():
    # Generate powerset of flag combinations. There are only 2^3 of them, so
    # we can reasonably expect that they are all are found.
    unseen_flags = {
        functools.reduce(operator.or_, flaglist, AFlag(0))
        for r in range(len(AFlag) + 1)
        for flaglist in itertools.combinations(AFlag, r)
    }

    @given(st.sampled_from(AFlag))
    def accept(flag):
        unseen_flags.discard(flag)

    accept()

    assert not unseen_flags


def test_flags_minimize_to_first_named_flag():
    assert minimal(st.sampled_from(LargeFlag)) == LargeFlag.bit0


def test_flags_minimizes_bit_count():
    assert (
        minimal(st.sampled_from(LargeFlag), lambda f: bit_count(f.value) > 1)
        == LargeFlag.bit0 | LargeFlag.bit1
    )


@pytest.mark.skipif(
    settings.get_current_profile_name() == "crosshair",
    reason="takes ~10 mins; path tree is too large",
)
def test_flags_finds_all_bits_set():
    assert find_any(st.sampled_from(LargeFlag), lambda f: f == ~LargeFlag(0))


def test_sample_unnamed_alias():
    assert find_any(st.sampled_from(UnnamedFlag), lambda f: f == UnnamedFlag.b)


def test_shrink_to_named_empty():
    assert minimal(st.sampled_from(UnnamedFlag)) == UnnamedFlag(0)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os
import sys
import warnings

from hypothesis import HealthCheck, settings

from tests.common.setup import run

if __name__ == "__main__":
    run()

    settings.register_profile(
        "default", settings(suppress_health_check=[HealthCheck.too_slow])
    )

    settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.django.toys.settings")

    # This triggers a deprecation warning on some older versions of Django
    # because of its use of the imp module.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        from django.core.management import execute_from_command_line

    try:
        from django.utils.deprecation import RemovedInDjango50Warning
    except ImportError:
        RemovedInDjango50Warning = ()

    try:
        from django.utils.deprecation import RemovedInDjango60Warning
    except ImportError:
        RemovedInDjango60Warning = ()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RemovedInDjango50Warning)
        warnings.simplefilter("ignore", category=RemovedInDjango60Warning)
        execute_from_command_line(sys.argv)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import re

import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import HypothesisWarning
from hypothesis.stateful import RuleBasedStateMachine, rule, run_state_machine_as_test


def test_error_is_in_finally():
    @given(st.data())
    def test(d):
        try:
            d.draw(st.lists(st.integers(), min_size=3, unique=True))
        finally:
            raise ValueError

    with pytest.raises(ValueError) as err:
        test()

    assert "[0, 1, -1]" in "\n".join(err.value.__notes__)


@given(st.data())
def test_warns_on_bool_strategy(data):
    with pytest.warns(
        HypothesisWarning,
        match=r"bool\(.+\) is always True, did you mean to draw a value\?",
    ):
        if st.booleans():  # 'forgot' to draw from the strategy
            pass


def test_adds_note_showing_which_strategy():
    class X:
        def __init__(self, y: int) -> None:
            assert y == 7

    @given(...)
    def inner(value: X):
        assert isinstance(value, X)

    rep = re.escape(repr(st.from_type(X)))
    with pytest.raises(AssertionError, match=f".*while generating 'value' from {rep}"):
        inner()


def test_adds_note_showing_which_strategy_stateful():
    strategy = st.integers().map(lambda x: x / 0)

    class Machine(RuleBasedStateMachine):
        @rule(value=strategy)
        def take_a_step(self, value):
            assert value

    msg = f"while generating 'value' from {strategy!r} for rule take_a_step"
    print(msg)
    with pytest.raises(ZeroDivisionError, match=f".*{re.escape(msg)}"):
        run_state_machine_as_test(Machine)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import uuid

import pytest

from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from tests.common.debug import assert_no_examples, check_can_generate_examples, find_any


def test_no_nil_uuid_by_default():
    assert_no_examples(st.uuids(), lambda x: x == uuid.UUID(int=0))


def test_can_generate_nil_uuid():
    find_any(st.uuids(allow_nil=True), lambda x: x == uuid.UUID(int=0))


def test_can_only_allow_nil_uuid_with_none_version():
    check_can_generate_examples(st.uuids(version=None, allow_nil=True))
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(st.uuids(version=4, allow_nil=True))
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(st.uuids(version=None, allow_nil="not a bool"))



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os

import pytest

import hypothesis
from hypothesis import errors
from hypothesis.internal import escalation as esc
from hypothesis.internal.compat import BaseExceptionGroup


def test_is_hypothesis_file_not_confused_by_prefix(monkeypatch):
    # Errors in third-party extensions such as `hypothesis-trio` or
    # `hypothesis-jsonschema` used to be incorrectly considered to be
    # Hypothesis internal errors, which could result in confusing error
    # messages. This test makes sure that files like:
    # `[...]/python3.12/site-packages/hypothesis_something/[...]`
    # are not considered as hypothesis files.
    root = os.path.dirname(hypothesis.__file__)
    assert esc.is_hypothesis_file(hypothesis.__file__)
    assert esc.is_hypothesis_file(esc.__file__)

    assert not esc.is_hypothesis_file(pytest.__file__)
    assert not esc.is_hypothesis_file(root + "-suffix")
    assert not esc.is_hypothesis_file(root + "-suffix/something.py")


@pytest.mark.parametrize("fname", ["", "<ipython-input-18-f7c304bea5eb>"])
def test_is_hypothesis_file_does_not_error_on_invalid_paths_issue_2319(fname):
    assert not esc.is_hypothesis_file(fname)


def test_multiplefailures_deprecation():
    with pytest.warns(errors.HypothesisDeprecationWarning):
        exc = errors.MultipleFailures
    assert exc is BaseExceptionGroup


def test_errors_attribute_error():
    with pytest.raises(AttributeError):
        errors.ThisIsNotARealAttributeDontCreateSomethingWithThisName


def test_handles_null_traceback():
    esc.InterestingOrigin.from_exception(Exception())


def test_handles_context():
    e = ValueError()
    e.__context__ = KeyError()
    origin = esc.InterestingOrigin.from_exception(e)
    assert "ValueError at " in str(origin)
    assert "  context: " in str(origin)
    assert "KeyError at " in str(origin)


def test_handles_groups():
    origin = esc.InterestingOrigin.from_exception(
        BaseExceptionGroup("message", [ValueError("msg2")])
    )
    assert "ExceptionGroup at " in str(origin)
    assert "child exception" in str(origin)
    assert "ValueError at " in str(origin)


def make_exceptions_with_cycles():
    err = ValueError()
    err.__context__ = err
    yield err

    err = TypeError()
    err.__context__ = BaseExceptionGroup("msg", [err])
    yield err

    inner = LookupError()
    err = BaseExceptionGroup("msg", [inner])
    inner.__context__ = err
    yield err

    inner = OSError()
    yield BaseExceptionGroup("msg", [inner, inner, inner])


@pytest.mark.parametrize("err", list(make_exceptions_with_cycles()))
def test_handles_cycles(err):
    esc.InterestingOrigin.from_exception(err)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import example, given, strategies as st
from hypothesis.errors import InvalidArgument

from tests.common.utils import fails_with


@example(False).via("Manually specified")
@given(st.booleans())
def test_ok_example_via(x):
    pass


def test_invalid_example_via():
    with pytest.raises(InvalidArgument):
        example(x=False).via(100)  # not a string!
    with pytest.raises(TypeError):
        example(x=False).via("abc", "def")  # too many args


@pytest.mark.parametrize(
    "kw",
    [
        {"condition": None},  # must be a bool
        {"reason": None},  # must be a string
        {"raises": None},  # not a BaseException (or even a type)
        {"raises": int},  # not a BaseException
        {"raises": [Exception]},  # not a tuple
        {"raises": (None,)},  # tuple containing a non-BaseException
        {"raises": ()},  # empty tuple doesn't make sense here
        # raising non-failure exceptions, eg KeyboardInterrupt, is tested below
    ],
    ids=repr,
)
def test_invalid_example_xfail_arguments(kw):
    with pytest.raises(InvalidArgument):
        example(x=False).xfail(**kw)


@example(True).xfail()
@example(True).xfail(reason="ignored for passing tests")
@example(True).xfail(raises=KeyError)
@example(True).xfail(raises=(KeyError, ValueError))
@example(True).xfail(True, reason="...")
@example(False).xfail(condition=False)
@given(st.none())
def test_many_xfail_example_decorators(fails):
    if fails:
        raise KeyError


@fails_with(AssertionError)
@example(x=True).xfail(raises=KeyError)
@given(st.none())
def test_xfail_reraises_non_specified_exception(x):
    assert not x


@fails_with(
    InvalidArgument,
    match=r"@example\(x=True\) raised an expected BaseException\('msg'\), "
    r"but Hypothesis does not treat this as a test failure",
)
@example(True).xfail()
@given(st.none())
def test_must_raise_a_failure_exception(x):
    if x:
        raise BaseException("msg")


@fails_with(
    AssertionError,
    match=r"Expected an exception from @example\(x=None\), but no exception was raised.",
)
@example(None).xfail()
@given(st.none())
def test_error_on_unexpected_pass_base(x):
    pass


@fails_with(
    AssertionError,
    match=r"Expected an AssertionError from @example\(x=None\), but no exception was raised.",
)
@example(None).xfail(raises=AssertionError)
@given(st.none())
def test_error_on_unexpected_pass_single(x):
    pass


@fails_with(
    AssertionError,
    match=r"Expected an AssertionError from @example\(x=None\), but no exception was raised.",
)
@example(None).xfail(raises=(AssertionError,))
@given(st.none())
def test_error_on_unexpected_pass_single_elem_tuple(x):
    pass


@fails_with(
    AssertionError,
    match=r"Expected a KeyError, or ValueError from @example\(x=None\), but no exception was raised.",
)
@example(None).xfail(raises=(KeyError, ValueError))
@given(st.none())
def test_error_on_unexpected_pass_multi(x):
    pass



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from functools import wraps

import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import InvalidArgument


def always_passes(*args, **kwargs):
    """Stand-in for a fixed version of an inner test.

    For example, pytest-trio would take the inner test, wrap it in an
    async-to-sync converter, and use the new func (not always_passes).
    """


@given(st.integers())
def test_can_replace_inner_test(x):
    raise AssertionError("This should be replaced")


test_can_replace_inner_test.hypothesis.inner_test = always_passes


def decorator(func):
    """An example of a common decorator pattern."""

    @wraps(func)
    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    return inner


@decorator
@given(st.integers())
def test_can_replace_when_decorated(x):
    raise AssertionError("This should be replaced")


test_can_replace_when_decorated.hypothesis.inner_test = always_passes


@pytest.mark.parametrize("x", [1, 2])
@given(y=st.integers())
def test_can_replace_when_parametrized(x, y):
    raise AssertionError("This should be replaced")


test_can_replace_when_parametrized.hypothesis.inner_test = always_passes


def test_can_replace_when_original_is_invalid():
    # Invalid: @given with too many positional arguments
    @given(st.integers(), st.integers())
    def invalid_test(x):
        raise AssertionError

    invalid_test.hypothesis.inner_test = always_passes

    # Even after replacing the inner test, calling the wrapper should still
    # fail.
    with pytest.raises(InvalidArgument, match="Too many positional arguments"):
        invalid_test()


def test_inner_is_original_even_when_invalid():
    def invalid_test(x):
        raise AssertionError

    original = invalid_test

    # Invalid: @given with no arguments
    invalid_test = given()(invalid_test)

    # Verify that the test is actually invalid
    with pytest.raises(
        InvalidArgument,
        match="given must be called with at least one argument",
    ):
        invalid_test()

    assert invalid_test.hypothesis.inner_test == original


def test_invokes_inner_function_with_args_by_name():
    # Regression test for https://github.com/HypothesisWorks/hypothesis/issues/3245
    @given(st.integers())
    def test(x):
        pass

    f = test.hypothesis.inner_test
    test.hypothesis.inner_test = wraps(f)(lambda **kw: f(**kw))
    test()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from unittest import TestCase

import pytest

from hypothesis import find, given, strategies as st
from hypothesis.errors import InvalidArgument
from hypothesis.stateful import RuleBasedStateMachine, rule


def test_cannot_use_without_a_runner():
    @given(st.runner())
    def f(x):
        pass

    with pytest.raises(InvalidArgument):
        f()


def test_cannot_use_in_find_without_default():
    with pytest.raises(InvalidArgument):
        find(st.runner(), lambda x: True)


def test_is_default_in_find():
    t = object()
    assert find(st.runner(default=t), lambda x: True) == t


@given(st.runner(default=1))
def test_is_default_without_self(runner):
    assert runner == 1


class TestStuff(TestCase):
    @given(st.runner())
    def test_runner_is_self(self, runner):
        assert runner is self

    @given(st.runner(default=3))
    def test_runner_is_self_even_with_default(self, runner):
        assert runner is self


class RunnerStateMachine(RuleBasedStateMachine):
    @rule(runner=st.runner())
    def step(self, runner):
        assert runner is self


TestState = RunnerStateMachine.TestCase



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from math import log
from random import Random

from hypothesis import (
    HealthCheck,
    Phase,
    Verbosity,
    assume,
    example,
    given,
    settings,
    strategies as st,
)
from hypothesis.internal.compat import ceil
from hypothesis.internal.conjecture.data import ConjectureData
from hypothesis.internal.conjecture.engine import ConjectureRunner

from tests.conjecture.common import interesting_origin


@st.composite
def problems(draw):
    m = draw(st.integers(256, 2**64 - 1))
    assume(m > 0)
    marker = draw(st.binary(max_size=8))
    bound = draw(st.integers(0, m - 1))
    return (m, marker, bound)


base_settings = settings(
    database=None,
    deadline=None,
    suppress_health_check=list(HealthCheck),
    max_examples=10,
    verbosity=Verbosity.normal,
    phases=(Phase.explicit, Phase.generate),
)


@example((4503599627370496, b"", 2861143707951135))
@example((88305152, b"%\x1b\xa0\xfa", 12394667))
@example((99742672384, b"\xf5|", 24300326997))
@example((1454610481571840, b"", 1076887621690235))
@example((15616, b"", 2508))
@example((65536, b"", 20048))
@example((256, b"", 0))
@example((512, b"", 258))
@example((2048, b"", 1792))
@example((3072, b"", 0))
@example((256, b"", 1))
@settings(
    base_settings,
    verbosity=Verbosity.normal,
    phases=(
        # We disable shrinking for this test because when it fails it's a sign
        # that the shrinker is working really badly, so it ends up very slow!
        Phase.explicit,
        Phase.generate,
    ),
    max_examples=20,
)
@given(problems())
def test_avoids_zig_zag_trap(p):
    m, marker, lower_bound = p
    n_bits = m.bit_length() + 1

    def test_function(data):
        m = data.draw_integer(0, 2**n_bits - 1)
        if m < lower_bound:
            data.mark_invalid()
        n = data.draw_integer(0, 2**n_bits - 1)
        if data.draw_bytes(len(marker), len(marker)) != marker:
            data.mark_invalid()
        if abs(m - n) == 1:
            data.mark_interesting(interesting_origin())

    runner = ConjectureRunner(
        test_function,
        database_key=None,
        settings=settings(base_settings, phases=(Phase.generate, Phase.shrink)),
        random=Random(0),
    )

    runner.cached_test_function((m, m + 1, marker))
    assert runner.interesting_examples
    runner.run()
    (v,) = runner.interesting_examples.values()
    data = ConjectureData.for_choices(v.choices)

    m = data.draw_integer(0, 2**n_bits - 1)
    n = data.draw_integer(0, 2**n_bits - 1)
    assert m == lower_bound
    if m == 0:
        assert n == 1
    else:
        assert n == m - 1

    budget = 2 * n_bits * ceil(log(n_bits, 2)) + 2
    assert runner.shrinks <= budget



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import os
import subprocess
import sys
from contextlib import contextmanager
from threading import RLock
from unittest import TestCase

import pytest

from hypothesis import example, given, strategies as st
from hypothesis._settings import (
    _CI_VARS,
    HealthCheck,
    Phase,
    Verbosity,
    default_variable,
    local_settings,
    settings,
)
from hypothesis.database import InMemoryExampleDatabase
from hypothesis.errors import (
    HypothesisDeprecationWarning,
    InvalidArgument,
)
from hypothesis.stateful import RuleBasedStateMachine, rule
from hypothesis.utils.conventions import not_set
from hypothesis.utils.deprecation import note_deprecation

from tests.common.utils import (
    checks_deprecated_behaviour,
    counts_calls,
    fails_with,
    skipif_emscripten,
    skipif_threading,
    validate_deprecation,
)

original_default = settings.get_profile("default").max_examples
_temp_register_profile_lock = RLock()


@contextmanager
def temp_register_profile(name, parent, **kwargs):
    with _temp_register_profile_lock:
        try:
            settings.register_profile(name, parent, **kwargs)
            yield
        finally:
            settings._profiles.pop(name)


_restore_profile_lock = RLock()


@contextmanager
def restore_profile():
    with _restore_profile_lock:
        # avoid polluting global state by resetting the loaded profile back to its
        # previous value in tests which use load_profile
        current_profile = settings.get_current_profile_name()
        try:
            yield
        finally:
            settings.load_profile(current_profile)


def test_cannot_set_non_settings():
    s = settings()
    with pytest.raises(AttributeError):
        s.databas_file = "some_file"


def test_settings_uses_defaults():
    s = settings()
    assert s.max_examples == settings.default.max_examples


def test_raises_attribute_error():
    with pytest.raises(AttributeError):
        settings().kittens


def test_respects_none_database():
    assert settings(database=None).database is None


def test_can_repeatedly_push_the_same_thing():
    s = settings(max_examples=12)
    t = settings(max_examples=17)
    original = settings().max_examples
    with local_settings(s):
        assert settings().max_examples == 12
        with local_settings(t):
            assert settings().max_examples == 17
            with local_settings(s):
                assert settings().max_examples == 12
                with local_settings(t):
                    assert settings().max_examples == 17
                assert settings().max_examples == 12
            assert settings().max_examples == 17
        assert settings().max_examples == 12
    assert settings().max_examples == original


def test_can_set_verbosity():
    settings(verbosity=Verbosity.quiet)
    settings(verbosity=Verbosity.normal)
    settings(verbosity=Verbosity.verbose)
    settings(verbosity=Verbosity.debug)


def test_can_not_set_verbosity_to_non_verbosity():
    with pytest.raises(InvalidArgument):
        settings(verbosity="kittens")


@pytest.mark.parametrize("db", [None, InMemoryExampleDatabase()])
def test_inherits_an_empty_database(db):
    with local_settings(settings(database=InMemoryExampleDatabase())):
        assert settings.default.database is not None
        s = settings(database=db)
        assert s.database is db
        with local_settings(s):
            t = settings()
        assert t.database is db


@pytest.mark.parametrize("db", [None, InMemoryExampleDatabase()])
def test_can_assign_database(db):
    x = settings(database=db)
    assert x.database is db


def test_will_reload_profile_when_default_is_absent():
    original = settings.default
    default_variable.value = None
    assert settings.default is original


def test_load_profile():
    with restore_profile():
        settings.load_profile("default")
        assert settings.default.max_examples == original_default
        assert settings.default.stateful_step_count == 50

        settings.register_profile(
            "test", settings(max_examples=10), stateful_step_count=5
        )
        settings.load_profile("test")

        assert settings.default.max_examples == 10
        assert settings.default.stateful_step_count == 5

        settings.load_profile("default")

        assert settings.default.max_examples == original_default
        assert settings.default.stateful_step_count == 50


def test_profile_names_must_be_strings():
    with pytest.raises(InvalidArgument):
        settings.register_profile(5)
    with pytest.raises(InvalidArgument):
        settings.get_profile(5)
    with pytest.raises(InvalidArgument):
        settings.load_profile(5)


def test_loading_profile_keeps_expected_behaviour():
    with restore_profile():
        settings.register_profile("ci", settings(max_examples=10000))
        settings.load_profile("ci")
        assert settings().max_examples == 10000
        with local_settings(settings(max_examples=5)):
            assert settings().max_examples == 5
        assert settings().max_examples == 10000


def test_load_non_existent_profile():
    with pytest.raises(InvalidArgument):
        settings.get_profile("nonsense")


def test_cannot_set_settings():
    x = settings()
    with pytest.raises(AttributeError):
        x.max_examples = "foo"
    with pytest.raises(AttributeError):
        x.database = "foo"
    assert x.max_examples != "foo"
    assert x.database != "foo"


def test_can_have_none_database():
    assert settings(database=None).database is None


@pytest.mark.parametrize("db", [None, InMemoryExampleDatabase()])
@pytest.mark.parametrize("bad_db", [":memory:", ".hypothesis/examples"])
def test_database_type_must_be_ExampleDatabase(db, bad_db):
    with local_settings(settings(database=db)):
        settings_property_db = settings.database
        with pytest.raises(InvalidArgument):
            settings(database=bad_db)
        assert settings.database is settings_property_db


def test_cannot_assign_default():
    with pytest.raises(AttributeError):
        settings.default = settings(max_examples=3)
    assert settings().max_examples != 3


@settings(max_examples=7)
@given(st.builds(lambda: settings.default))
def test_settings_in_strategies_are_from_test_scope(s):
    assert s.max_examples == 7


TEST_SETTINGS_ALONE = """
from hypothesis import settings
from hypothesis.strategies import integers

@settings()
def test_settings_alone():
    pass
"""


# runpytest_inprocess uses invalidate_caches in pytest, which is not thread safe
# (I presume; produces keyerrors).
@skipif_threading
def test_settings_alone(pytester):
    # Disable cacheprovider, since we don't need it and it's flaky on pyodide
    script = pytester.makepyfile(TEST_SETTINGS_ALONE)
    result = pytester.runpytest_inprocess(script, "-p", "no:cacheprovider")
    out = "\n".join(result.stdout.lines)
    msg = "Using `@settings` on a test without `@given` is completely pointless."
    assert msg in out
    assert "InvalidArgument" in out
    assert result.ret == 1


@fails_with(InvalidArgument)
def test_settings_applied_twice_is_error():
    @given(st.integers())
    @settings()
    @settings()
    def test_nothing(x):
        pass


@settings()
@given(st.integers())
def test_outer_ok(x):
    pass


@given(st.integers())
@settings()
def test_inner_ok(x):
    pass


def test_settings_as_decorator_must_be_on_callable():
    with pytest.raises(InvalidArgument):
        settings()(1)


ASSERT_DATABASE_PATH = """
import tempfile
from hypothesis import settings
from hypothesis.configuration import set_hypothesis_home_dir
from hypothesis.database import DirectoryBasedExampleDatabase

settings.load_profile("default")
settings.default.database

if __name__ == '__main__':
    new_home = tempfile.mkdtemp()
    set_hypothesis_home_dir(new_home)
    db = settings.default.database
    assert isinstance(db, DirectoryBasedExampleDatabase), db
    assert db.path.is_relative_to(new_home), (db.path, new_home)
"""


@skipif_emscripten
def test_puts_the_database_in_the_home_dir_by_default(tmp_path):
    script = tmp_path / "assertlocation.py"
    script.write_text(ASSERT_DATABASE_PATH, encoding="utf-8")
    subprocess.check_call([sys.executable, str(script)])


def test_database_is_reference_preserved():
    s = settings(database=not_set)

    assert s.database is s.database


@settings(verbosity=Verbosity.verbose)
@example(x=99)
@given(st.integers())
def test_settings_apply_for_explicit_examples(x):
    # Regression test for #1521
    assert settings.default.verbosity == Verbosity.verbose


class TestGivenExampleSettingsExplicitCalled(TestCase):
    """Real nasty edge case here.

    in #2160, if ``example`` is after ``given`` but before ``settings``,
    it will be completely ignored.

    If we set phases to only ``explicit``, the test case will never be called!

    We have to run an assertion outside of the test case itself.
    """

    @counts_calls
    def call_target(self):
        pass

    @given(st.booleans())
    @example(True)
    @settings(phases=[Phase.explicit])
    # counts_calls is not thread safe (modifying global f.calls attr)
    @skipif_threading
    def test_example_explicit(self, x):
        self.call_target()

    def tearDown(self):
        # In #2160, this is 0.
        assert self.call_target.calls == 1


def test_setattr_on_settings_singleton_is_error():
    # https://github.com/pandas-dev/pandas/pull/22679#issuecomment-420750921
    # Should be setting attributes on settings.default, not settings!
    with pytest.raises(AttributeError):
        settings.max_examples = 10


def test_deadline_given_none():
    x = settings(deadline=None).deadline
    assert x is None


def test_deadline_given_valid_int():
    x = settings(deadline=1000).deadline
    assert isinstance(x, datetime.timedelta)
    assert x.days == 0
    assert x.seconds == 1
    assert x.microseconds == 0


def test_deadline_given_valid_float():
    x = settings(deadline=2050.25).deadline
    assert isinstance(x, datetime.timedelta)
    assert x.days == 0
    assert x.seconds == 2
    assert x.microseconds == 50250


def test_deadline_given_valid_timedelta():
    x = settings(deadline=datetime.timedelta(days=1, microseconds=15030000)).deadline
    assert isinstance(x, datetime.timedelta)
    assert x.days == 1
    assert x.seconds == 15
    assert x.microseconds == 30000


@pytest.mark.parametrize("value", ["always"])
def test_can_not_set_print_blob_to_non_print_settings(value):
    with pytest.raises(InvalidArgument):
        settings(print_blob=value)


settings_step_count = 1


@settings(stateful_step_count=settings_step_count)
class StepCounter(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.step_count = 0

    @rule()
    def count_step(self):
        self.step_count += 1

    def teardown(self):
        assert self.step_count <= settings_step_count


test_settings_decorator_applies_to_rule_based_state_machine_class = StepCounter.TestCase


def test_two_settings_decorators_applied_to_state_machine_class_raises_error():
    with pytest.raises(InvalidArgument):

        @settings()
        @settings()
        class StatefulTest(RuleBasedStateMachine):
            pass


def test_settings_decorator_applied_to_non_state_machine_class_raises_error():
    with pytest.raises(InvalidArgument):

        @settings()
        class NonStateMachine:
            pass


def test_assigning_to_settings_attribute_on_state_machine_raises_error():
    class StateMachine(RuleBasedStateMachine):
        @rule(x=st.none())
        def a_rule(self, x):
            assert x is None

    with pytest.raises(AttributeError):
        StateMachine.settings = settings()

    state_machine_instance = StateMachine()
    state_machine_instance.settings = "any value"


def test_derandomise_with_explicit_database_is_invalid():
    with pytest.raises(InvalidArgument):
        settings(derandomize=True, database=InMemoryExampleDatabase())


@pytest.mark.parametrize(
    "kwargs",
    [
        {"max_examples": -1},
        {"max_examples": 2.5},
        {"stateful_step_count": -1},
        {"stateful_step_count": 2.5},
        {"deadline": -1},
        {"deadline": 0},
        {"deadline": -0.7},
        {"deadline": 86400000000000000.2},
        {"deadline": datetime.timedelta(microseconds=-1)},
        {"deadline": datetime.timedelta(0)},
        {"deadline": True},
        {"deadline": False},
        {"backend": "nonexistent_backend"},
        {"suppress_health_check": ["nonexistent_healthcheck"]},
        {"phases": ["nonexistent_phase"]},
        {"phases": 0},
        {"verbosity": -1},
        {"verbosity": "nonexistent_verbosity"},
    ],
)
def test_invalid_settings_are_errors(kwargs):
    with pytest.raises(InvalidArgument):
        settings(**kwargs)


def test_invalid_parent():
    class NotSettings:
        def __repr__(self):
            return "(not settings repr)"

    not_settings = NotSettings()

    with pytest.raises(InvalidArgument, match=r"parent=\(not settings repr\)"):
        settings(not_settings)


def test_default_settings_do_not_use_ci():
    assert settings.get_profile("default").suppress_health_check == ()


def test_show_changed():
    s = settings(settings.get_profile("default"), max_examples=999, database=None)
    assert s.show_changed() == "database=None, max_examples=999"


def test_note_deprecation_checks_date():
    with pytest.warns(HypothesisDeprecationWarning) as rec:
        note_deprecation("This is bad", since="RELEASEDAY", has_codemod=False)
    assert len(rec) == 1
    with pytest.raises(AssertionError):
        note_deprecation("This is way too old", since="1999-12-31", has_codemod=False)


def test_note_deprecation_checks_has_codemod():
    with pytest.warns(
        HypothesisDeprecationWarning,
        match="The `hypothesis codemod` command-line tool",
    ):
        note_deprecation("This is bad", since="2021-01-01", has_codemod=True)


def test_deprecated_settings_warn_on_set_settings():
    with validate_deprecation():
        settings(suppress_health_check=[HealthCheck.return_value])
    with validate_deprecation():
        settings(suppress_health_check=[HealthCheck.not_a_test_method])


@checks_deprecated_behaviour
def test_deprecated_settings_not_in_settings_all_list():
    al = HealthCheck.all()
    ls = list(HealthCheck)
    assert al == ls
    assert HealthCheck.return_value not in ls
    assert HealthCheck.not_a_test_method not in ls


@skipif_emscripten
def test_check_defaults_to_derandomize_when_running_on_ci():
    env = dict(os.environ)
    env["CI"] = "true"

    assert (
        subprocess.check_output(
            [
                sys.executable,
                "-c",
                "from hypothesis import settings\nprint(settings().derandomize)",
            ],
            env=env,
            text=True,
            encoding="utf-8",
        ).strip()
        == "True"
    )


@skipif_emscripten
def test_check_defaults_to_randomize_when_not_running_on_ci():
    env = dict(os.environ)
    for key in _CI_VARS:
        env.pop(key, None)

    assert (
        subprocess.check_output(
            [
                sys.executable,
                "-c",
                "from hypothesis import settings\nprint(settings().derandomize)",
            ],
            env=env,
            text=True,
            encoding="utf-8",
        ).strip()
        == "False"
    )


@skipif_threading  # modifying global state (profiles) during testing
def test_reloads_the_loaded_profile_if_registered_again():
    with restore_profile():
        test_profile = "some nonsense profile purely for this test"
        test_value = 123456
        settings.register_profile(test_profile, settings(max_examples=test_value))
        settings.load_profile(test_profile)
        assert settings.default.max_examples == test_value
        test_value_2 = 42
        settings.register_profile(test_profile, settings(max_examples=test_value_2))
        assert settings.default.max_examples == test_value_2


CI_TESTING_SCRIPT = """
from hypothesis import settings

if __name__ == '__main__':
    settings.register_profile("ci", settings(max_examples=42))
    assert settings.default.max_examples == 42
"""


@skipif_emscripten
def test_will_automatically_pick_up_changes_to_ci_profile_in_ci():
    env = dict(os.environ)
    env["CI"] = "true"
    subprocess.check_call(
        [sys.executable, "-c", CI_TESTING_SCRIPT],
        env=env,
        text=True,
        encoding="utf-8",
    )


def test_register_profile_avoids_intermediate_profiles():
    parent = settings()
    s = settings(parent, max_examples=10)
    with temp_register_profile("for_intermediate_test", s):
        assert settings.get_profile("for_intermediate_test")._fallback is parent


@checks_deprecated_behaviour
@settings(max_examples=10)
@given(st.integers())
def test_cannot_register_profile_from_inside_test(x):
    settings.register_profile("problematic", settings(max_examples=20))


def test_can_set_verbosity_to_strings():
    assert settings(verbosity="quiet").verbosity is Verbosity.quiet
    assert settings(verbosity="normal").verbosity is Verbosity.normal
    assert settings(verbosity="verbose").verbosity is Verbosity.verbose
    assert settings(verbosity="debug").verbosity is Verbosity.debug


def test_can_set_phase_to_strings():
    assert settings(phases=["reuse"]).phases == (Phase.reuse,)
    assert settings(phases=["reuse", "explicit"]).phases == (
        Phase.explicit,
        Phase.reuse,
    )


def test_can_set_suppressions_to_strings():
    assert settings(
        suppress_health_check=["filter_too_much"]
    ).suppress_health_check == (HealthCheck.filter_too_much,)
    assert settings(
        suppress_health_check=["filter_too_much", "too_slow"]
    ).suppress_health_check == (HealthCheck.filter_too_much, HealthCheck.too_slow)


def test_verbosity_is_comparable():
    assert Verbosity.quiet < Verbosity.normal
    assert Verbosity.quiet <= Verbosity.quiet
    assert Verbosity.quiet == Verbosity.quiet
    assert Verbosity.quiet >= Verbosity.quiet
    assert Verbosity.debug > Verbosity.quiet

    # make sure we're comparing by int value, not by str value
    assert Verbosity.quiet < Verbosity.normal < Verbosity.verbose < Verbosity.debug

    # also comparable with other ints
    assert Verbosity.quiet < 1
    assert Verbosity.quiet <= 1
    assert Verbosity.quiet == 0
    assert Verbosity.quiet >= 0
    assert Verbosity.normal > 0


@checks_deprecated_behaviour
def test_can_set_verbosity_to_integers():
    assert Verbosity(0) is Verbosity.quiet
    assert Verbosity(1) is Verbosity.normal
    assert Verbosity(2) is Verbosity.verbose
    assert Verbosity(3) is Verbosity.debug


@checks_deprecated_behaviour
def test_can_set_phase_to_integers():
    assert Phase(0) is Phase.explicit
    assert Phase(1) is Phase.reuse
    assert Phase(2) is Phase.generate
    assert Phase(4) is Phase.shrink


@checks_deprecated_behaviour
def test_can_set_suppressions_to_integers():
    assert HealthCheck(1) is HealthCheck.data_too_large
    assert HealthCheck(2) is HealthCheck.filter_too_much
    assert HealthCheck(3) is HealthCheck.too_slow


def test_invalid_integer_phase_raises():
    with pytest.raises(ValueError):
        Phase(99)


def test_invalid_integer_healthcheck_raises():
    with pytest.raises(ValueError):
        HealthCheck(99)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import ast
import inspect
import subprocess
import sys
import textwrap
from types import ModuleType

import pytest

from hypothesis import example, given, note, settings, strategies as st
from hypothesis.internal.compat import PYPY
from hypothesis.internal.constants_ast import (
    Constants,
    ConstantVisitor,
    TooManyConstants,
    _constants_file_str,
    _constants_from_source,
    constants_from_module,
    is_local_module_file,
)

from tests.common.utils import skipif_emscripten, skipif_threading

constant_ints = st.integers(max_value=-101) | st.integers(min_value=101)
constant_floats = st.floats(allow_nan=False, allow_infinity=False)
constant_bytes = st.binary(min_size=1, max_size=50)
constant_strings = st.text(min_size=1, max_size=10).filter(lambda s: not s.isspace())
constants = constant_ints | constant_floats | constant_bytes | constant_strings

constants_classes = st.builds(
    Constants,
    integers=st.sets(constant_ints),
    floats=st.sets(constant_floats),
    bytes=st.sets(constant_bytes),
    strings=st.sets(constant_strings),
)


def constants_from_ast(tree):
    visitor = ConstantVisitor(limit=True)
    visitor.visit(tree)
    return visitor.constants


def test_constants_set_from_type_invalid():
    with pytest.raises(ValueError):
        Constants().set_for_type("not_a_type")


@given(st.integers())
def test_constants_contains(n):
    assert n in Constants(integers={n})


@given(constants_classes)
def test_constants_not_equal_to_set(constants):
    assert constants != set()
    assert constants != set(constants)


@pytest.mark.parametrize(
    "source, expected",
    [
        (
            """
            a1 = 142
            a2 = 3.14
            a3 = 'test1'
            a4 = b'test2'
            a5 = (101, 102)
            a6 = frozenset([103])
            """,
            {142, 3.14, 101, 102, 103, "test1", b"test2"},
        ),
        (
            "a = (101, (102, 103), frozenset([104, 105]))",
            {101, 102, 103, 104, 105},
        ),
        ("a = {'b': 101}", {"b", 101}),
        ("a = [101]", {101}),
        ("a = +142", {142}),
        ("a = 101 + 102", {101, 102}),
        ("a = ~ 142", {142}),
        # the following cases are ignored:
        # * booleans
        # * math.inf and math.nan (not constants, but we don't want to collect them
        #   even if they were)
        # * f-strings
        # * long strings
        # * pure-whitespace strings
        # * standalone string expressions (strings not assigned to a variable).
        #   This covers docstrings of all kinds.
        # * small integers
        # * the empty bytestring b""
        ("a = True", set()),
        ("a = False", set()),
        ("a = not False", set()),
        ("a = 1e999", set()),
        ("a = math.inf", set()),
        ("a = math.nan", set()),
        ('a = f"test {x}"', set()),
        (f'a = "{"b" * 100}"', set()),
        ('a = ""', set()),
        ('a = " "', set()),
        ('a = "\\n    \\n  \\n"', set()),
        ("'test'", set()),
        ("'test with \\n newlines'", set()),
        ("a = 10", set()),
        ("a = -1", set()),
        ("a = b''", set()),
    ],
)
def test_constants_from_ast(source, expected):
    source = textwrap.dedent(source)
    tree = ast.parse(source)
    assert set(constants_from_ast(tree)) == expected


@given(st.integers(max_value=-101))
def test_parses_negatives(n):
    assert constants_from_ast(ast.parse(f"a = {n}")) == Constants(integers={n})


@given(st.tuples(constants))
def test_tuple_constants(value):
    tree = ast.parse(str(value))
    assert set(constants_from_ast(tree)) == set(value)


@given(st.frozensets(constants))
def test_frozenset_constants(value):
    tree = ast.parse(str(value))
    assert set(constants_from_ast(tree)) == set(value)


@skipif_threading
@skipif_emscripten
@pytest.mark.xfail(
    condition=settings.get_current_profile_name() != "ci",
    strict=False,
    reason="Requires clean environment",
)
def test_constants_from_running_file(tmp_path):
    p = tmp_path / "my_constants.py"
    p.write_text(
        textwrap.dedent(
            """
        import sys
        # stdlib
        import json
        # third-party
        import pytest
        import hypothesis
        from hypothesis.internal.constants_ast import (
            is_local_module_file,
            constants_from_module,
            Constants
        )

        # these modules are in fact detected as local if they are installed
        # as editable (as is common for contributors). Prevent the ast constant
        # logic from picking up on them
        for module in sys.modules.copy():
            if module.startswith("hypofuzz"):
                del sys.modules[module]

        constants = Constants()
        for module in sys.modules.values():
            if getattr(module, "__file__", None) is not None and is_local_module_file(
                module.__file__
            ):
                constants |= constants_from_module(module)

        expected = Constants(
            strings={'float', 'string', 'bytes', 'integer', 'test', 'hypofuzz', '__file__'},
            floats={3.14},
            bytes={b'test'},
            integers={142, 101, 102, 103, 104},
        )
        assert constants == expected, set(constants).symmetric_difference(set(expected))

        # local
        a = 142
        b = "test"
        c = True
        d = 3.14
        e = b"test"
        f = (101, 102)
        g = frozenset([103, 104])
        """,
        ),
        encoding="utf-8",
    )
    # this test doubles as a regression test for
    # https://github.com/HypothesisWorks/hypothesis/issues/4375. Fail on comparisons
    # between bytes and str.
    subprocess.check_call([sys.executable, "-bb", str(p)])


def test_constants_from_bad_module():
    # covering test for the except branch
    module = ModuleType("nonexistent")
    assert constants_from_module(module) == Constants()


@pytest.mark.parametrize(
    "path",
    [
        "/path/to/tests/module",
        "/path/to/test/module",
        "/a/test_file.py",
        "/a/file_test.py",
    ],
)
def test_local_modules_ignores_test_modules(path):
    assert not is_local_module_file(path)


@skipif_threading
@pytest.mark.skipif(PYPY, reason="no memory error on pypy")
def test_ignores_ast_parse_error(tmp_path):
    p = tmp_path / "errors_on_parse.py"
    p.write_text("[1, " * 200 + "]" * 200, encoding="utf-8")
    module = ModuleType("<test_ignores_ast_parse_error>")
    module.__file__ = str(p)

    source = inspect.getsource(module)
    with pytest.raises(MemoryError):
        ast.parse(source)

    assert constants_from_module(module) == Constants()


@example(
    constants=Constants(
        integers={-101, 101},
        floats={101.0},
        bytes=set(),
        strings=set(),
    ),
)
@example(
    constants=Constants(
        integers={-101, 100},
        floats={-101.0, 101.0},
        bytes={b"-101.0", b"101.0"},
        strings={"-101.0", "101.0"},
    ),
)
@given(constants_classes)
def test_constant_visitor_roundtrips_string(constants):
    # our files in storage_directory("constants") rely on this roundtrip
    visitor = ConstantVisitor(limit=True)

    s = _constants_file_str(constants)
    visitor.visit(ast.parse(s))

    note(f"{constants=}")
    note(f"{visitor.constants=}")
    assert visitor.constants == constants


def test_too_many_constants():
    visitor = ConstantVisitor(limit=True)
    # start at n=1000 to avoid ConstantVisitor ignoring small integers
    s = "; ".join(
        f"n = {i}" for i in range(1000, 1000 + ConstantVisitor.CONSTANTS_LIMIT + 1)
    )
    # visitor should raise on too many constants
    with pytest.raises(TooManyConstants):
        visitor.visit(ast.parse(s))

    # and also _constants_from_source should return empty on too many constants
    assert _constants_from_source(s, limit=True) == Constants()

    # but it parses fine with limit=False
    visitor = ConstantVisitor(limit=False)
    visitor.visit(ast.parse(s))
    assert _constants_from_source(s, limit=False) == Constants(
        integers=set(range(1000, 1000 + ConstantVisitor.CONSTANTS_LIMIT + 1))
    )


@skipif_threading  # concurrent writes to the same file
def test_module_too_large(tmp_path):
    constant = 11231783

    p = tmp_path / "large_file.py"
    content = f"a = {constant}\n\n" + "#" * (512 * 1024 + 1)
    p.write_text(content, encoding="utf-8")

    module = ModuleType("large_module")
    module.__file__ = str(p)
    assert constants_from_module(module) == Constants()
    assert constants_from_module(module, limit=False) == Constants(integers={constant})



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import typing
import warnings

import numpy as np
import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import SmallSearchSpaceWarning
from hypothesis.extra.numpy import ArrayLike, NDArray, _NestedSequence, _SupportsArray

from .test_from_dtype import STANDARD_TYPES
from tests.common.debug import assert_simple_property, find_any

STANDARD_TYPES_TYPE = [dtype.type for dtype in STANDARD_TYPES]

needs_np_typing = {"reason": "numpy.typing is not available"}
needs_np_private_typing = {"reason": "numpy._typing is not available"}


@given(dtype=st.from_type(np.dtype))
def test_resolves_dtype_type(dtype):
    assert isinstance(dtype, np.dtype)


def test_does_not_resolve_nonscalar_types():
    # this was previously a parametrized test over np.object_ and np.void which
    # used the same repr code path for the test. But then numpy changed their types
    # such that we defer evaluation for st.from_type and are no longer identical
    # to st.builds, but rather something morally equivalent to it. So we have
    # these slightly more complicated checks.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", SmallSearchSpaceWarning)
        assert_simple_property(st.from_type(np.object_), lambda value: value is None)

    with pytest.raises(TypeError):
        # np.void() requires an argument, and so throws when instantiated
        assert_simple_property(st.from_type(np.void))


@pytest.mark.parametrize("typ", STANDARD_TYPES_TYPE)
def test_resolves_and_varies_numpy_scalar_type(typ):
    # Check that we find an instance that is not equal to the default
    x = find_any(st.from_type(typ), lambda x: x != type(x)())
    assert isinstance(x, typ)


@pytest.mark.parametrize("atype", [np.ndarray, NDArray])
def test_resolves_unspecified_array_type(atype):
    if atype is not None:
        assert_simple_property(st.from_type(atype), lambda v: isinstance(v, np.ndarray))


def workaround(dtype):
    # Total hack to work around https://github.com/numpy/numpy/issues/24043
    if np.__version__ == "1.25.0" and dtype == np.dtype("bytes").type:
        return pytest.param(dtype, marks=[pytest.mark.xfail(strict=False)])
    return dtype


# https://numpy.org/devdocs/release/1.22.0-notes.html#ndarray-dtype-and-number-are-now-runtime-subscriptable
@pytest.mark.skipif(
    tuple(int(x) for x in np.__version__.split(".")[:2]) < (1, 22), reason="see comment"
)
@pytest.mark.parametrize("typ", [workaround(t) for t in STANDARD_TYPES_TYPE])
def test_resolves_specified_ndarray_type(typ):
    assert_simple_property(
        st.from_type(np.ndarray[typ]),
        lambda arr: isinstance(arr, np.ndarray) and arr.dtype.type == typ,
    )

    assert_simple_property(
        st.from_type(np.ndarray[typing.Any, typ]),
        lambda arr: isinstance(arr, np.ndarray) and arr.dtype.type == typ,
    )


@pytest.mark.skipif(NDArray is None, **needs_np_typing)
@pytest.mark.parametrize("typ", [workaround(t) for t in STANDARD_TYPES_TYPE])
def test_resolves_specified_NDArray_type(typ):
    assert_simple_property(
        st.from_type(NDArray[typ]),
        lambda arr: isinstance(arr, np.ndarray) and arr.dtype.type == typ,
    )


@pytest.mark.skipif(NDArray is None, **needs_np_typing)
def test_resolves_NDArray_with_dtype_union():
    strat = st.from_type(NDArray[np.float64 | np.complex128])
    find_any(strat, lambda arr: arr.dtype == np.dtype("float64"))
    find_any(strat, lambda arr: arr.dtype == np.dtype("complex128"))


@pytest.mark.skipif(ArrayLike is None, **needs_np_typing)
@given(arr_like=st.from_type(ArrayLike))
def test_resolves_ArrayLike_type(arr_like):
    arr = np.array(arr_like)
    assert isinstance(arr, np.ndarray)
    # The variation is too large to assert anything else about arr, but the
    # ArrayLike contract just says that it can be coerced into an array (which
    # we just did).


@pytest.mark.skipif(_NestedSequence is None, **needs_np_private_typing)
def test_resolves_specified_NestedSequence():
    @given(seq=st.from_type(_NestedSequence[int]))
    def test(seq):
        assert hasattr(seq, "__iter__")

        def flatten(lst):
            for el in lst:
                try:
                    yield from flatten(el)
                except TypeError:
                    yield el

        assert all(isinstance(i, int) for i in flatten(seq))

    test()


@pytest.mark.skipif(_NestedSequence is None, **needs_np_private_typing)
@given(seq=st.from_type(_NestedSequence))
def test_resolves_unspecified_NestedSequence(seq):
    assert hasattr(seq, "__iter__")


@pytest.mark.skipif(_SupportsArray is None, **needs_np_private_typing)
@given(arr=st.from_type(_SupportsArray))
def test_resolves_unspecified_SupportsArray(arr):
    assert hasattr(arr, "__array__")


@pytest.mark.skipif(_SupportsArray is None, **needs_np_private_typing)
def test_resolves_SupportsArray():
    @given(arr=st.from_type(_SupportsArray[int]))
    def test(arr):
        assert hasattr(arr, "__array__")
        assert np.asarray(arr).dtype.kind == "i"

    test()


@pytest.mark.skipif(
    _NestedSequence is None or _SupportsArray is None, **needs_np_private_typing
)
def test_resolve_ArrayLike_equivalent():
    # This is the current (1.24.3) definition of ArrayLike,
    # with problematic parts commented out.
    ArrayLike_like = (
        _SupportsArray
        # | _NestedSequence[_SupportsArray],
        | bool
        | int
        | float
        | complex
        | str
        | bytes
        | _NestedSequence[
            bool | int | float | complex | str,
            # | bytes,
        ]
    )

    @given(arr_like=st.from_type(ArrayLike_like))
    def test(arr_like):
        arr = np.array(arr_like)
        assert isinstance(arr, np.ndarray)

    test()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import dataclasses
import math
import sys
from contextlib import contextmanager
from random import Random
from threading import RLock

import pytest

from hypothesis import HealthCheck, Phase, settings, strategies as st
from hypothesis.control import current_build_context, currently_in_test_context
from hypothesis.internal.conjecture import engine as engine_module
from hypothesis.internal.conjecture.choice import ChoiceNode, ChoiceT
from hypothesis.internal.conjecture.data import ConjectureData, Status
from hypothesis.internal.conjecture.engine import ConjectureRunner
from hypothesis.internal.conjecture.provider_conformance import (
    choice_types_constraints,
    constraints_strategy,
)
from hypothesis.internal.conjecture.providers import COLLECTION_DEFAULT_MAX_SIZE
from hypothesis.internal.conjecture.utils import calc_label_from_name
from hypothesis.internal.escalation import InterestingOrigin
from hypothesis.internal.floats import SMALLEST_SUBNORMAL
from hypothesis.internal.intervalsets import IntervalSet

SOME_LABEL = calc_label_from_name("some label")


def interesting_origin(n: int | None = None) -> InterestingOrigin:
    """
    Creates and returns an InterestingOrigin, parameterized by n, such that
    interesting_origin(n) == interesting_origin(m) iff n = m.

    Since n=None may by chance concide with an explicitly-passed value of n, I
    recommend not mixing interesting_origin() and interesting_origin(n) in the
    same test.
    """
    try:
        int("not an int")
    except Exception as e:
        origin = InterestingOrigin.from_exception(e)
        return dataclasses.replace(origin, lineno=n if n is not None else origin.lineno)


def run_to_data(f):
    runner = ConjectureRunner(
        f,
        settings=settings(
            max_examples=300, database=None, suppress_health_check=list(HealthCheck)
        ),
        random=Random(0),
    )
    runner.run()
    assert runner.interesting_examples
    (last_data,) = runner.interesting_examples.values()
    return last_data


def run_to_nodes(f):
    return run_to_data(f).nodes


_buffer_size_lock = RLock()


@contextmanager
def buffer_size_limit(n):
    with _buffer_size_lock:
        original = engine_module.BUFFER_SIZE
        try:
            engine_module.BUFFER_SIZE = n
            yield
        finally:
            engine_module.BUFFER_SIZE = original


def shrinking_from(start):
    def accept(f):
        runner = ConjectureRunner(
            f,
            settings=settings(
                max_examples=5000,
                database=None,
                suppress_health_check=list(HealthCheck),
                # avoid running the explain phase in shrinker.shrink() in tests
                # which don't test the inquisitor.
                phases=set(settings.default.phases) - {Phase.explain},
            ),
            random=Random(0),
        )
        runner.cached_test_function(start)
        assert runner.interesting_examples

        (last_data,) = runner.interesting_examples.values()
        return runner.new_shrinker(last_data, lambda d: d.status == Status.INTERESTING)

    return accept


def fresh_data(*, random=None, observer=None) -> ConjectureData:
    context = current_build_context() if currently_in_test_context() else None
    if context is not None and settings().backend == "crosshair":
        # we should reeaxmine fresh_data sometime and see if we can replace it
        # with nicer and higher level hypothesis idioms.
        #
        # For now it doesn't work well with crosshair tests. This is no big
        # loss, because these tests often rely on hypothesis-provider-specific
        # things.
        pytest.skip(
            "Fresh data is too low level (and too much of a hack) to be "
            "worth supporting when testing with crosshair"
        )

    if random is None:
        if context is None:
            # ensure usage of fresh_data() is not flaky outside of property tests.
            raise ValueError(
                "must pass a seeded Random instance to fresh_data() when "
                "outside of a build context"
            ) from None

        # within property tests, ensure fresh_data uses a controlled source of
        # randomness.
        # drawing this from the current build context is almost *too* magical. But
        # the alternative is an extra @given(st.randoms()) everywhere we use
        # fresh_data, so eh.

        # @example uses a zero-length data, which means we can't use a
        # hypothesis-backed random (which would entail drawing from the data).
        # In this case, use a deterministic Random(0).
        random = (
            context.data.draw(st.randoms(use_true_random=True))
            if (choices := context.data.max_choices) is None or choices > 0
            else Random(0)
        )

    return ConjectureData(random=random, observer=observer)


def clamped_shrink_towards(constraints):
    v = constraints["shrink_towards"]
    if constraints["min_value"] is not None:
        v = max(constraints["min_value"], v)
    if constraints["max_value"] is not None:
        v = min(constraints["max_value"], v)
    return v


def draw_value(choice_type, constraints):
    data = fresh_data()
    return getattr(data, f"draw_{choice_type}")(**constraints)


@st.composite
def choices(draw):
    (choice_type, constraints) = draw(choice_types_constraints())
    return draw_value(choice_type, constraints)


@st.composite
def nodes(draw, *, was_forced=None, choice_types=None):
    if choice_types is None:
        (choice_type, constraints) = draw(choice_types_constraints())
    else:
        choice_type = draw(st.sampled_from(choice_types))
        constraints = draw(constraints_strategy(choice_type))
    # choice nodes don't include forced in their constraints. see was_forced attribute
    del constraints["forced"]
    value = draw_value(choice_type, constraints)
    was_forced = draw(st.booleans()) if was_forced is None else was_forced

    return ChoiceNode(
        type=choice_type, value=value, constraints=constraints, was_forced=was_forced
    )


def nodes_inline(*values: list[ChoiceT]) -> list[ChoiceNode]:
    """
    For inline-creating a choice node or list of choice nodes, where you don't
    care about the constraints. This uses maximally-permissable constraints and
    infers the choice_type you meant based on the type of the value.

    You can optionally pass (value, constraints) to as an element in order to override
    the default constraints for that element.
    """
    mapping = {
        float: (
            "float",
            {
                "min_value": -math.inf,
                "max_value": math.inf,
                "allow_nan": True,
                "smallest_nonzero_magnitude": SMALLEST_SUBNORMAL,
            },
        ),
        int: (
            "integer",
            {
                "min_value": None,
                "max_value": None,
                "weights": None,
                "shrink_towards": 0,
            },
        ),
        str: (
            "string",
            {
                "intervals": IntervalSet(((0, sys.maxunicode),)),
                "min_size": 0,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
            },
        ),
        bytes: ("bytes", {"min_size": 0, "max_size": COLLECTION_DEFAULT_MAX_SIZE}),
        bool: ("boolean", {"p": 0.5}),
    }
    nodes = []
    for value in values:
        override_constraints = {}
        if isinstance(value, tuple):
            (value, override_constraints) = value
            if override_constraints is None:
                override_constraints = {}

        (choice_type, constraints) = mapping[type(value)]

        nodes.append(
            ChoiceNode(
                type=choice_type,
                value=value,
                constraints=constraints | override_constraints,
                was_forced=False,
            )
        )

    return tuple(nodes)


def float_constr(
    min_value=-math.inf,
    max_value=math.inf,
    *,
    allow_nan=True,
    smallest_nonzero_magnitude=SMALLEST_SUBNORMAL,
):
    return {
        "min_value": float(min_value),
        "max_value": float(max_value),
        "allow_nan": allow_nan,
        "smallest_nonzero_magnitude": float(smallest_nonzero_magnitude),
    }


def integer_constr(min_value=None, max_value=None, *, weights=None, shrink_towards=0):
    return {
        "min_value": min_value,
        "max_value": max_value,
        "weights": weights,
        "shrink_towards": shrink_towards,
    }


def string_constr(intervals, *, min_size=0, max_size=COLLECTION_DEFAULT_MAX_SIZE):
    return {"intervals": intervals, "min_size": min_size, "max_size": max_size}


# we could in theory define bytes_constr and boolean_constr, but without any
# default kw values they aren't really a time save.



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from random import Random

import pytest

from hypothesis import (
    HealthCheck,
    Verbosity,
    core,
    given,
    settings,
    strategies as st,
)

from tests.common.utils import Why, no_shrink, xfail_on_crosshair


@pytest.mark.skipif(
    settings.get_current_profile_name() == "crosshair",
    reason="we do not yet pass backends the global random seed, so they are not deterministic",
)
def test_seeds_off_internal_random():
    choices1 = []
    choices2 = []

    @given(st.integers())
    def f1(n):
        choices1.append(n)

    @given(st.integers())
    def f2(n):
        choices2.append(n)

    core.threadlocal._hypothesis_global_random = Random(0)
    state = core.threadlocal._hypothesis_global_random.getstate()
    f1()

    core.threadlocal._hypothesis_global_random.setstate(state)
    f2()

    assert choices1 == choices2


@xfail_on_crosshair(Why.nested_given)
def test_nesting_with_control_passes_health_check():
    @given(st.integers(0, 100), st.random_module())
    @settings(
        max_examples=5,
        database=None,
        deadline=None,
        suppress_health_check=[HealthCheck.nested_given],
    )
    def test_blah(x, rnd):
        @given(st.integers())
        @settings(
            max_examples=100, phases=no_shrink, database=None, verbosity=Verbosity.quiet
        )
        def test_nest(y):
            assert y < x

        with pytest.raises(AssertionError):
            test_nest()

    test_blah()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import hypothesis.strategies as st
from hypothesis.internal.conjecture.data import ConjectureData
from hypothesis.strategies._internal.strategies import FilteredStrategy


def test_filter_iterations_are_marked_as_discarded():
    variable_equal_to_zero = 0  # non-local references disables filter-rewriting
    x = st.integers().filter(lambda x: x == variable_equal_to_zero)

    data = ConjectureData.for_choices([1, 0])
    assert data.draw(x) == 0
    assert data.has_discards


def test_filtered_branches_are_all_filtered():
    s = FilteredStrategy(st.integers() | st.text(), (bool,))
    assert all(isinstance(x, FilteredStrategy) for x in s.branches)


def test_filter_conditions_may_be_empty():
    s = FilteredStrategy(st.integers(), conditions=())
    s.condition(0)


def test_nested_filteredstrategy_flattens_conditions():
    s = FilteredStrategy(
        FilteredStrategy(st.text(), conditions=(bool,)),
        conditions=(len,),
    )
    assert s.filtered_strategy is st.text()
    assert s.flat_conditions == (bool, len)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import time
from unittest import TestCase

import pytest

from hypothesis import (
    Phase,
    Verbosity,
    assume,
    example,
    given,
    note,
    reporting,
    settings,
)
from hypothesis.errors import DeadlineExceeded, HypothesisWarning, InvalidArgument
from hypothesis.internal.compat import ExceptionGroup
from hypothesis.strategies import floats, integers, text

from tests.common.utils import (
    assert_falsifying_output,
    capture_out,
    fails_with,
    skipif_threading,
)


class TestInstanceMethods(TestCase):
    @given(integers())
    @example(1)
    def test_hi_1(self, x):
        assert isinstance(x, int)

    @given(integers())
    @example(x=1)
    def test_hi_2(self, x):
        assert isinstance(x, int)

    @given(x=integers())
    @example(x=1)
    def test_hi_3(self, x):
        assert isinstance(x, int)


def test_kwarg_example_on_testcase():
    class Stuff(TestCase):
        @given(integers())
        @example(x=1)
        def test_hi(self, x):
            assert isinstance(x, int)

    Stuff("test_hi").test_hi()


def test_errors_when_run_with_not_enough_args():
    @given(integers(), int)
    @example(1)
    def foo(x, y):
        pass

    with pytest.raises(TypeError):
        foo()


def test_errors_when_run_with_not_enough_kwargs():
    @given(integers(), int)
    @example(x=1)
    def foo(x, y):
        pass

    with pytest.raises(TypeError):
        foo()


def test_can_use_examples_after_given():
    long_str = "This is a very long string that you've no chance of hitting"

    @example(long_str)
    @given(text())
    def test_not_long_str(x):
        assert x != long_str

    with pytest.raises(AssertionError):
        test_not_long_str()


def test_can_use_examples_before_given():
    long_str = "This is a very long string that you've no chance of hitting"

    @given(text())
    @example(long_str)
    def test_not_long_str(x):
        assert x != long_str

    with pytest.raises(AssertionError):
        test_not_long_str()


def test_can_use_examples_around_given():
    long_str = "This is a very long string that you've no chance of hitting"
    short_str = "Still no chance"

    seen = []

    @example(short_str)
    @given(text())
    @example(long_str)
    def test_not_long_str(x):
        seen.append(x)

    test_not_long_str()
    assert set(seen[:2]) == {long_str, short_str}


@pytest.mark.parametrize(("x", "y"), [(1, False), (2, True)])
@example(z=10)
@given(z=integers())
def test_is_a_thing(x, y, z):
    pass


def test_no_args_and_kwargs():
    with pytest.raises(InvalidArgument):
        example(1, y=2)


def test_no_empty_examples():
    with pytest.raises(InvalidArgument):
        example()


def test_does_not_print_on_explicit_examples_if_no_failure():
    @example(1)
    @given(integers())
    def test_positive(x):
        assert x > 0

    with (
        reporting.with_reporter(reporting.default),
        pytest.raises(AssertionError),
        capture_out() as out,
    ):
        test_positive()
    out = out.getvalue()
    assert "Falsifying example: test_positive(1)" not in out


def test_prints_output_for_explicit_examples():
    @example(-1)
    @given(integers())
    def test_positive(x):
        assert x > 0

    assert_falsifying_output(test_positive, "Falsifying explicit", x=-1)


def test_prints_verbose_output_for_explicit_examples():
    @settings(verbosity=Verbosity.verbose)
    @example("NOT AN INTEGER")
    @given(integers())
    def test_always_passes(x):
        pass

    assert_falsifying_output(
        test_always_passes,
        expected_exception=None,
        example_type="Trying explicit",
        x="NOT AN INTEGER",
    )


def test_captures_original_repr_of_example():
    @example(x=[])
    @given(integers())
    def test_mutation(x):
        x.append(1)
        assert not x

    assert_falsifying_output(test_mutation, "Falsifying explicit", x=[])


def test_examples_are_tried_in_order():
    @example(x=1)
    @example(x=2)
    @given(integers())
    @settings(phases=[Phase.explicit])
    @example(x=3)
    def test(x):
        print(f"x -> {x}")

    with capture_out() as out, reporting.with_reporter(reporting.default):
        test()
    ls = out.getvalue().splitlines()
    assert ls == ["x -> 1", "x -> 2", "x -> 3"]


def test_prints_note_in_failing_example():
    @example(x=42)
    @example(x=43)
    @given(integers())
    def test(x):
        note(f"x -> {x}")
        assert x == 42

    with pytest.raises(AssertionError) as err:
        test()
    assert "x -> 43" in err.value.__notes__
    assert "x -> 42" not in err.value.__notes__


def test_must_agree_with_number_of_arguments():
    @example(1, 2)
    @given(integers())
    def test(a):
        pass

    with pytest.raises(InvalidArgument):
        test()


@skipif_threading  # deadline disabled under threading
@fails_with(DeadlineExceeded)
@example(10)
@settings(phases=[Phase.explicit], deadline=1)
@given(integers())
def test(x):
    time.sleep(10)


@given(value=floats(0, 1))
@example(value=0.56789)
@pytest.mark.parametrize("threshold", [0.5, 1])
def test_unsatisfied_assumption_during_explicit_example(threshold, value):
    # Regression test, expected to pass / skip depending on parametrize.
    # See https://github.com/HypothesisWorks/hypothesis/issues/2125
    assume(value < threshold)


@pytest.mark.parametrize("exc", [ExceptionGroup, AssertionError])
def test_multiple_example_reporting(exc):
    @example(1)
    @example(2)
    @settings(report_multiple_bugs=exc is ExceptionGroup, phases=[Phase.explicit])
    @given(integers())
    def inner_test_multiple_failing_examples(x):
        assert x < 2
        assert x < 1

    with pytest.raises(exc):
        inner_test_multiple_failing_examples()


def test_simplifies_multiple_examples_with_same_error():
    # When multiple examples fail with the same error, only the simplest is shown
    @example(x=1000)
    @example(x=1)
    @example(x=100)
    @settings(report_multiple_bugs=True, phases=[Phase.explicit])
    @given(x=integers())
    def test_fn(x):
        assert x < 0

    with pytest.raises(AssertionError) as exc_info:
        test_fn()

    err = exc_info.value
    # Should show note about other examples
    assert any(
        "2 other explicit examples also failed" in note for note in err.__notes__
    )
    # Should show the simplest example (x=1 is shorter than x=100 or x=1000)
    assert any("x=1," in note for note in err.__notes__)


def test_shows_all_examples_at_verbose():
    # At verbose, all examples with same error should be shown
    @example(x=1000)
    @example(x=1)
    @settings(
        report_multiple_bugs=True, phases=[Phase.explicit], verbosity=Verbosity.verbose
    )
    @given(x=integers())
    def test_fn(x):
        assert x < 0

    with pytest.raises(ExceptionGroup) as exc_info:
        test_fn()

    group = exc_info.value
    # Both examples should be shown
    assert len(group.exceptions) == 2
    # No simplification note should be present
    for exc in group.exceptions:
        assert not any(
            "other explicit example" in note for note in getattr(exc, "__notes__", [])
        )


def test_different_errors_not_simplified():
    # Examples with different errors should all be shown
    @example(x=1)
    @example(x=2)
    @settings(report_multiple_bugs=True, phases=[Phase.explicit])
    @given(x=integers())
    def test_fn(x):
        if x == 1:
            raise ValueError("one")
        raise TypeError("two")

    with pytest.raises(ExceptionGroup) as exc_info:
        test_fn()

    assert {type(e) for e in exc_info.value.exceptions} == {ValueError, TypeError}


@example(text())
@given(text())
def test_example_decorator_accepts_strategies(s):
    """The custom error message only happens when the test has already failed."""


def test_helpful_message_when_example_fails_because_it_was_passed_a_strategy():
    @example(text())
    @given(text())
    def t(s):
        assert isinstance(s, str)

    try:
        t()
    except HypothesisWarning as err:
        assert isinstance(err.__cause__, AssertionError)
    else:
        raise NotImplementedError("should be unreachable")


def test_stop_silently_dropping_examples_when_decorator_is_applied_to_itself():
    def f():
        pass

    test = example("outer")(example("inner"))(f)
    assert len(test.hypothesis_explicit_examples) == 2



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import example, given, strategies as st
from hypothesis.errors import InvalidArgument

# Be aware that tests in this file pass strategies as arguments to @example.
# That's normally a mistake, but for these tests it's intended.
# If one of these tests fails, Hypothesis will complain about the
# @example/strategy interaction, but it should be safe to ignore that
# error message and focus on the underlying failure instead.

base_reusable_strategies = (
    st.text(),
    st.binary(),
    st.dates(),
    st.times(),
    st.timedeltas(),
    st.booleans(),
    st.complex_numbers(),
    st.floats(),
    st.floats(-1.0, 1.0),
    st.integers(),
    st.integers(1, 10),
    st.integers(1),
    # Note that `just` and `sampled_from` count as "reusable" even if their
    # values are mutable, because the user has implicitly promised that they
    # don't care about the same mutable value being returned by separate draws.
    st.just([]),
    st.sampled_from([[]]),
    st.tuples(st.integers()),
)


@st.deferred
def reusable():
    """Meta-strategy that produces strategies that should have
    ``.has_reusable_values == True``."""
    return st.one_of(
        # This looks like it should be `one_of`, but `sampled_from` is correct
        # because we want this meta-strategy to yield strategies as its values.
        st.sampled_from(base_reusable_strategies),
        # This sometimes produces invalid combinations of arguments, which
        # we filter out further down with an explicit validation check.
        st.builds(
            st.floats,
            min_value=st.none() | st.floats(allow_nan=False),
            max_value=st.none() | st.floats(allow_nan=False),
            allow_infinity=st.booleans(),
            allow_nan=st.booleans(),
        ),
        st.builds(st.just, st.builds(list)),
        st.builds(st.sampled_from, st.lists(st.builds(list), min_size=1)),
        st.lists(reusable).map(st.one_of),
        st.lists(reusable).map(lambda ls: st.tuples(*ls)),
    )


def is_valid(s):
    try:
        s.validate()
        return True
    except InvalidArgument:
        return False


reusable = reusable.filter(is_valid)

assert not reusable.is_empty


def many_examples(examples):
    """Helper decorator to apply the ``@example`` decorator multiple times,
    once for each given example."""

    def accept(f):
        for e in examples:
            f = example(e)(f)
        return f

    return accept


@many_examples(base_reusable_strategies)
@many_examples(st.tuples(s) for s in base_reusable_strategies)
@given(reusable)
def test_reusable_strategies_are_all_reusable(s):
    assert s.has_reusable_values


@many_examples(base_reusable_strategies)
@given(reusable)
def test_filter_breaks_reusability(s):
    cond = True

    def nontrivial_filter(x):
        """Non-trivial filtering function, intended to remain opaque even if
        some strategies introspect their filters."""
        return cond

    assert s.has_reusable_values
    assert not s.filter(nontrivial_filter).has_reusable_values


@many_examples(base_reusable_strategies)
@given(reusable)
def test_map_breaks_reusability(s):
    cond = True

    def nontrivial_map(x):
        """Non-trivial mapping function, intended to remain opaque even if
        some strategies introspect their mappings."""
        if cond:
            return x
        else:
            return None

    assert s.has_reusable_values
    assert not s.map(nontrivial_map).has_reusable_values


@many_examples(base_reusable_strategies)
@given(reusable)
def test_flatmap_breaks_reusability(s):
    cond = True

    def nontrivial_flatmap(x):
        """Non-trivial flat-mapping function, intended to remain opaque even
        if some strategies introspect their flat-mappings."""
        if cond:
            return st.just(x)
        else:
            return st.none()

    assert s.has_reusable_values
    assert not s.flatmap(nontrivial_flatmap).has_reusable_values


@pytest.mark.parametrize(
    "strat",
    [
        st.lists(st.booleans()),
        st.sets(st.booleans()),
        st.dictionaries(st.booleans(), st.booleans()),
    ],
)
def test_mutable_collections_do_not_have_reusable_values(strat):
    assert not strat.has_reusable_values


def test_recursion_does_not_break_reusability():
    x = st.deferred(lambda: st.none() | st.tuples(x))
    assert x.has_reusable_values



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import django
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models


class Company(models.Model):
    name = models.CharField(max_length=100, unique=True)


class Store(models.Model):
    name = models.CharField(max_length=100, unique=True)
    company = models.ForeignKey(Company, null=False, on_delete=models.CASCADE)


class Car(models.Model):
    # See https://github.com/HypothesisWorks/hypothesis/issues/2369
    model = models.CharField(max_length=100, unique=True)


class CharmField(models.Field):
    def db_type(self, connection):
        return "char(1)"


class CustomishField(models.Field):
    def db_type(self, connection):
        return "char(1)"


class Customish(models.Model):
    customish = CustomishField()


class Customer(models.Model):
    name = models.CharField(max_length=100, unique=True)
    email = models.EmailField(max_length=100, unique=True)
    gender = models.CharField(max_length=50, null=True)  # noqa  # avoid nullable strs
    age = models.IntegerField()
    birthday = models.DateTimeField()


class Charming(models.Model):
    charm = CharmField()


class CouldBeCharming(models.Model):
    charm = CharmField(null=True)


class SelfLoop(models.Model):
    me = models.ForeignKey("self", null=True, on_delete=models.SET_NULL)


class LoopA(models.Model):
    b = models.ForeignKey("LoopB", null=False, on_delete=models.CASCADE)


class LoopB(models.Model):
    a = models.ForeignKey("LoopA", null=True, on_delete=models.SET_NULL)


class ManyNumerics(models.Model):
    i1 = models.IntegerField()
    i2 = models.SmallIntegerField()
    i3 = models.BigIntegerField()

    p1 = models.PositiveIntegerField()
    p2 = models.PositiveSmallIntegerField()

    d = models.DecimalField(decimal_places=2, max_digits=5)


class ManyTimes(models.Model):
    time = models.TimeField()
    date = models.DateField()
    duration = models.DurationField()


class OddFields(models.Model):
    uuid = models.UUIDField()
    slug = models.SlugField()
    url = models.URLField()
    ipv4 = models.GenericIPAddressField(protocol="IPv4")
    ipv6 = models.GenericIPAddressField(protocol="IPv6")


class CustomishDefault(models.Model):
    customish = CustomishField(default="b")


class FileFields(models.Model):
    file1 = models.FileField()


class MandatoryComputed(models.Model):
    name = models.CharField(max_length=100, unique=True)
    company = models.ForeignKey(Company, null=False, on_delete=models.CASCADE)

    def __init__(self, **kw):
        if "company" in kw:
            raise RuntimeError
        cname = kw["name"] + "_company"
        kw["company"] = Company.objects.create(name=cname)
        super().__init__(**kw)


def validate_even(value):
    if value % 2 != 0:
        raise ValidationError("")


class RestrictedFields(models.Model):
    text_field_4 = models.TextField(max_length=4, blank=True)
    char_field_4 = models.CharField(max_length=4, blank=True)
    choice_field_text = models.TextField(choices=(("foo", "Foo"), ("bar", "Bar")))
    choice_field_int = models.IntegerField(choices=((1, "First"), (2, "Second")))
    null_choice_field_int = models.IntegerField(
        choices=((1, "First"), (2, "Second")), null=True, blank=True
    )
    choice_field_grouped = models.TextField(
        choices=(
            ("Audio", (("vinyl", "Vinyl"), ("cd", "CD"))),
            ("Video", (("vhs", "VHS Tape"), ("dvd", "DVD"))),
            ("unknown", "Unknown"),
        )
    )
    even_number_field = models.IntegerField(validators=[validate_even])
    non_blank_text_field = models.TextField(blank=False)


class SelfModifyingField(models.IntegerField):
    def pre_save(self, model_instance, add):
        value = getattr(model_instance, self.attname)
        value += 1
        setattr(model_instance, self.attname, value)
        return value


class CompanyExtension(models.Model):
    company = models.OneToOneField(Company, primary_key=True, on_delete=models.CASCADE)

    self_modifying = SelfModifyingField()


class UserSpecifiedAutoId(models.Model):
    my_id = models.AutoField(primary_key=True)


if django.VERSION >= (5, 0, 0):
    import math

    class Pizza(models.Model):
        AREA = math.pi * models.F("radius") ** 2

        radius = models.IntegerField(validators=[MinValueValidator(1)])
        slices = models.PositiveIntegerField(validators=[MinValueValidator(2)])
        total_area = models.GeneratedField(
            expression=AREA,
            output_field=models.FloatField(),
            db_persist=True,
        )
        slice_area = models.GeneratedField(
            expression=AREA / models.F("slices"),
            output_field=models.FloatField(),
            db_persist=False,
        )



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import strategies as st
from hypothesis.internal.conjecture.data import ConjectureData
from hypothesis.strategies._internal.lazy import LazyStrategy


def test_labels_are_cached():
    x = st.integers()
    assert x.label is x.label


def test_labels_are_distinct():
    assert st.integers().label != st.text().label


@st.composite
def foo(draw):
    return draw(st.none())


@st.composite
def bar(draw):
    return draw(st.none())


@st.composite
def baz(draw):
    return draw(st.booleans())


def test_equivalent_composites_have_same_label():
    assert foo().label == bar().label


def test_different_composites_have_different_labels():
    assert foo().label != baz().label


def test_one_of_label_is_distinct():
    a = st.integers()
    b = st.booleans()
    assert st.one_of(a, b).label != st.one_of(b, a).label


def test_lists_label_by_element():
    assert st.lists(st.integers()).label != st.lists(st.booleans()).label


def test_label_of_deferred_strategy_is_well_defined():
    recursive = st.deferred(lambda: st.lists(recursive))
    recursive.label


@pytest.mark.parametrize(
    "strategy",
    [
        lambda: [st.none()],
        lambda: [st.integers()],
        lambda: [st.lists(st.floats())],
        lambda: [st.none(), st.integers(), st.lists(st.floats())],
    ],
)
def test_sampled_from_label_with_strategies_does_not_change(strategy):
    s1 = st.sampled_from(strategy())
    s2 = st.sampled_from(strategy())
    assert s1.label == s2.label


def test_label_of_enormous_sampled_range():
    # this should not take forever.
    st.sampled_from(range(2**30)).label


@pytest.mark.parametrize(
    "strategy",
    [
        lambda: st.deferred(lambda: st.integers()),
        lambda: LazyStrategy(st.integers, (), {}),
    ],
)
def test_draw_uses_wrapped_label(strategy):
    cd = ConjectureData.for_choices([0])
    strategy = strategy()
    cd.draw(strategy)
    cd.freeze()

    assert len(cd.spans) == 2
    assert cd.spans[1].label == st.integers().label


def test_deferred_label():
    strategy = st.deferred(lambda: st.integers())
    assert strategy.label != st.integers().label
